/*  dvdisaster: Additional error correction for optical media.
 *  Copyright (C) 2004,2005 Carsten Gnoerlich.
 *  Project home page: http://www.dvdisaster.com
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA,
 *  or direct your browser at http://www.gnu.org.
 */

#include "dvdisaster.h"

#include <time.h>

/***
 *** Debugging functions.
 ***/

/*
 * Debugging function to seed the image with random correctable errors
 */

void RandomError(char *prefix)
{  ImageInfo *ii;
   gint64 block_idx[255];
   gint64 s,si;
   int block_sel[255];
   int i,percent,last_percent = 0;
   int n_data,n_eras;
   double eras_scale, blk_scale;

   srandom((int)time(NULL));

   if(!Closure->redundancy || !strcmp(Closure->redundancy, "normal")) n_eras = 32; 
   else if(!strcmp(Closure->redundancy, "high")) n_eras = 64;
   else n_eras = atoi(Closure->redundancy);

   n_data = 255-n_eras;
   eras_scale = (n_eras+1)/((double)RAND_MAX+1.0);
   blk_scale = (double)n_data/((double)RAND_MAX+1.0);


   /*** Open the image file */

   ii = OpenImage(prefix, OPEN_WRITEABLE, NULL);

   /*** Setup block pointers */

   s = (ii->sectors+n_data-1)/n_data;

   for(si=0, i=0; i<n_data; si+=s, i++)
     block_idx[i] = si;

   PrintMsg(_("\nGenerating random correctable erasures (max = %d).\n"),n_eras);

   /*** Randomly delete the blocks */

   for(si=0; si<s; si++)
   {  int n_erasures = (int)(eras_scale*(double)random());

      /* Reset the block selector */

      for(i=0; i<n_data; i++)  
	block_sel[i] = 0;

      /* Randomly pick n blocks */

      for(i=0; i<n_erasures; i++)
      {  int idx;
      
         do
	 {  idx = (int)(blk_scale*(double)random());
	 } while(block_sel[idx]);

	 block_sel[idx] = 1;
      }

      /* Delete the randomly picked blocks */

      for(i=0; i<n_data; i++)
      {  if(block_sel[i] && block_idx[i]<ii->sectors)
	  {  if(lseek(ii->file, (gint64)(2048*block_idx[i]), SEEK_SET) == -1)
	      Stop(_("Could not seek to image sector %lld:\n%s\n"),block_idx[i],strerror(errno));
	     if(write(ii->file, Closure->deadSector, 2048) != 2048)
	       Stop(_("Could not write image sector %lld:\n%s\n"),block_idx[i],strerror(errno));
	  }

          block_idx[i]++;
      }

      percent = (100*si)/s;
      if(last_percent != percent) 
      {  PrintMsg(_("Progress: %3d%%\b\b\b\b\b\b\b\b\b\b\b\b\b\b"),percent);
	 last_percent = percent;
      }
   }

   PrintMsg(_("Progress: 100%%\n"
	    "Recover the image using the --fix option before doing another --random-errors run.\n"
	    "Otherwise you'll accumulate >= %d erasures/ECC block and the image will be lost.\n"), n_eras);

   FreeImageInfo(ii);
}

/*
 * Debugging function to simulate images with single
 * byte errors (except for faulty cabling and/or controllers,
 * this should never happen)
 */

void Byteset(char *prefix, char *arg)
{  ImageInfo *ii;
   gint64 s;
   int i,byte;
   char *cpos = NULL;

   /*** Open the image and map files */

   ii = OpenImage(prefix, OPEN_WRITEABLE, NULL);

   /*** See which byte to set */

   cpos = strchr(arg,',');
   if(!cpos) Stop(_("2nd argument is missing"));
   *cpos = 0;

   s = atoi(arg);
   arg = cpos+1;

   cpos = strchr(arg,',');
   if(!cpos) Stop(_("3rd argument is missing"));
   *cpos = 0;

   i = atoi(arg);
   byte = atoi(cpos+1);

   if(s<0 || s>ii->sectors-1)
     Stop(_("Sector must be in range [0..%lld]"),ii->sectors-1);

   if(i<0 || i>=2048)
     Stop(_("Byte position must be in range [0..2047]"));

   if(byte<0 || byte>=256)
     Stop(_("Byte value must be in range [0..255]"));

   PrintMsg(_("Setting byte %d in sector %lld to value %d.\n"), i, s, byte); 

   /*** Set the byte */
   
   s = 2048*s + i;
   
   if(lseek(ii->file, (gint64)s, SEEK_SET) == -1)
     Stop(_("Could not seek to first image sector:\n%s\n"),strerror(errno));

   if(write(ii->file, &byte, 1) == -1)
     Stop(_("Could not write the new byte value"));

   FreeImageInfo(ii);
}

/*
 * Debugging function to simulate medium with unreadable sectors
 */

void Erase(char *prefix, char *arg)
{  ImageInfo *ii;
   gint64 start,end,s;
   char *dashpos = NULL;

   /*** Open the image file */

   ii = OpenImage(prefix, OPEN_WRITEABLE, NULL);

   /*** See which sectors to erase */

   dashpos = strchr(arg,'-');
   if(dashpos)
   {  *dashpos = 0;
      start = atoi(arg);
      end  = atoi(dashpos+1);
   }
   else start = end = atoi(arg);

   if(start>end || start < 0 || end >= ii->sectors)
     Stop(_("Sectors must be in range [0,%lld].\n"),ii->sectors-1);

   PrintMsg(_("Erasing sectors [%lld,%lld]\n"),start,end);

   /*** Erase them. */

   if(lseek(ii->file, (gint64)(2048*start), SEEK_SET) == -1)
     Stop(_("Could not seek to first image sector:\n%s\n"),strerror(errno));

   for(s=start; s<=end; s++)
   {  int m = (end == ii->sectors-1) ? ii->inLast : 2048;
      int n = write(ii->file, Closure->deadSector, m);

      if(n != m)
	Stop(_("Could not write image sector %lld:\n%s\n"),s,strerror(errno));
   }

   /*** Clean up */

   FreeImageInfo(ii);
}

/*
 * Just count the number of erasures.
 */

void Dryrun(char *imagename)
{  ImageInfo *ii; 
   Method *method = FindMethod(Closure->methodName); 

   if(!method) Stop(_("\nMethod %s not available.\n"
		     "Use -m without parameters for a method list.\n"), 
	            Closure->methodName);

   ii = OpenImage(imagename, 0, NULL);

   method->dryrun(ii);

   FreeImageInfo(ii);
}
