/*  dvdisaster: Additional error correction for optical media.
 *  Copyright (C) 2004,2005 Carsten Gnoerlich.
 *  Project home page: http://www.dvdisaster.com
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA,
 *  or direct your browser at http://www.gnu.org.
 */

#include "dvdisaster.h"

#ifdef SYS_MINGW
 #include <windows.h>
 #include <winnls.h>
 #include <tlhelp32.h>
 #include <psapi.h>
#endif

//#define QUIET

/*
 * A special marker for a unread sector 
 */

static void prepare_dead_sector()
{   char *end_marker;
    int end_length; 

    Closure->deadSector = g_malloc0(2048);

    g_sprintf(Closure->deadSector,
	      "dvdisaster dead sector marker\n"
	      "This sector could not be read from the image.\n"
	      "Its contents have been substituted by the dvdisaster read routine.\n");

    end_marker = "dvdisaster dead sector end marker\n";
    end_length = strlen(end_marker);
    memcpy(Closure->deadSector+2046-end_length, end_marker, end_length); 
}

/*
 * See if we can handle the specified .ecc file
 */

int VerifyVersion(EccHeader *eh, int fatal)
{  
   if(Closure->version < eh->neededVersion)
   {  if(fatal)
       Stop(_("\n* This .ecc file requires dvdisaster V%d.%d.%d or higher.\n"
	      "* Please visit http://www.dvdisaster.com for an upgrade."),
	    eh->neededVersion/10000,
	    (eh->neededVersion%10000)/100,
	    eh->neededVersion%100);

      return 1;
   }

   return 0;
}

/*
 * This is a transitional wrapper for upward compatibility
 * with version 0.60 and beyond
 */

void CalculateFootprint(ImageInfo *ii, EccHeader *eh)
{  struct MD5Context md5ctxt;
   unsigned char buf[2048];
   int n;

   MD5Init(&md5ctxt);
   lseek(ii->file, 2048*eh->fpSector, SEEK_SET);
   n = read(ii->file, buf, 2048);
   if(n != 2048)
     Stop(_("could not read first image sector (only %d bytes):\n%s"),n,strerror(errno));
   MD5Update(&md5ctxt, buf, n);
   MD5Final(ii->mediumFP, &md5ctxt);
   if(!memcmp(buf,Closure->deadSector,2048))  /* No sector, no md5sum */
     memset(ii->mediumFP, 0 ,16);
}

/***
 *** Open the image file and fill in the ImageInfo struct.
 ***/

#define CRCBUFSIZE (1024*256)

#if 0     /* Do we really need this? */
int xread(int file, char *buf, int len)
{  int n,count=0;

   while(len) 
   {  n = read(file, buf, len);
      if(n<0) return n;        /* read error */
      if(n==0) return count;   /* end of file */
      count+=n;
      len-=n;
      buf+=n;
   }

   return count;               /* everything read okay */
}
#endif

ImageInfo* OpenImage(char *imagename, int mode, FILE *ecc)
{  ImageInfo *ii;
   EccHeader eh; 
   void *buf;
   guint32 crcbuf[CRCBUFSIZE];
   int crcidx = 0;
   struct MD5Context md5ctxt;
   struct stat mystat; 
   int file_flags,n;
   gint64 eh_sectors = 0;

   /*** Basic calculations */

   ii = g_malloc0(sizeof(ImageInfo));
   ii->name = imagename;

   buf = g_malloc(2048);

   /*** Try opening the image file. */

   if(!(mode & VERBOSE_FULLSCAN))
     PrintMsg(_("\nOpening %s"),imagename);

   if(stat(imagename, &mystat) == -1)
     Stop(_(": not present."));

   file_flags = mode & OPEN_WRITEABLE ? O_RDWR : O_RDONLY;
#ifdef HAVE_O_LARGEFILE
   file_flags |= O_LARGEFILE; 
#endif
#ifdef SYS_MINGW
   file_flags |= O_BINARY;
#endif
   if((ii->file = open(imagename, file_flags, S_IRWXU)) == -1)
     Stop(": %s", strerror(errno));

   ii->size = mystat.st_size;
   CalcSectors(mystat.st_size, &ii->sectors, &ii->inLast);
   if(!(mode & VERBOSE_FULLSCAN))
     PrintMsg(_(": %lld medium sectors.\n"), ii->sectors);


   if(mode & CHECK_CRC32)
   {   rewind(ecc);
       fread(&eh, sizeof(EccHeader), 1, ecc);
       eh_sectors = uchar_to_gint64(eh.sectors);
   }     

   if(mode & (CREATE_CRC32 | CHECK_CRC32) )
   {   if(fseeko(ecc, (gint64)sizeof(EccHeader), SEEK_SET) == -1)
          Stop(_("fseek failed skipping ecc header: %s"),strerror(errno));

       crcidx = mode & CREATE_CRC32 ? 0 : CRCBUFSIZE;
       MD5Init(&ii->md5Ctxt);
   }

   /*** Calculate md5sum of first sector */

   MD5Init(&md5ctxt);
   n = read(ii->file, buf, 2048);
   if(n != 2048)
     Stop(_("could not read first image sector (only %d bytes):\n%s"),n,strerror(errno));
   MD5Update(&md5ctxt, buf, n);
   MD5Final(ii->mediumFP, &md5ctxt);
   if(!memcmp(buf,Closure->deadSector,2048))  /* No sector, no md5sum */
     memset(ii->mediumFP, 0 ,16);
   lseek(ii->file, 0, SEEK_SET);     /* rewind, need to read that again */

   /*** If a full scan is requested, calculate the image md5sum
	and count the missing sectors and crc failures. */

   if(mode & OPEN_FULLSCAN)
   {  struct MD5Context md5ctxt;
      gint64 s, first_missing, last_missing;
      int last_percent,current_missing;
      char *msg;
      
      if(mode & VERBOSE_FULLSCAN)
	   msg = _("- testing sectors  : %3d%%\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b");
      else msg = _("Scanning image sectors: %3d%%\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b");

      last_percent = 0;
      ii->sectorsMissing  = 0;
      first_missing = last_missing = -1;

      MD5Init(&md5ctxt);
      for(s=0; s<ii->sectors; s++)
      {  int n,percent;

         /* Read the next sector */

         n = read(ii->file, buf, 2048);
	 if(n != 2048 && (s != ii->sectors - 1 || n != ii->inLast))
	   Stop(_("premature end in image (only %d bytes):\n"),n,strerror(errno));

	 /* Look for the dead sector marker */

	 current_missing = !memcmp(buf, Closure->deadSector, n);
	 if(current_missing)
	 {  if(first_missing < 0) first_missing = s;
	    last_missing = s;
	    ii->sectorsMissing++;
	 }

	 /* Report dead sectors. Combine subsequent missing sectors into one report. */

	 if(mode & VERBOSE_FULLSCAN)
	   if(!current_missing || s==ii->sectors-1)
	   {  if(first_missing>=0)
	       {  if(first_missing == last_missing)
		       PrintLog(_("* missing sector   : %lld\n"), first_missing);
		  else PrintLog(_("* missing sectors  : %lld - %lld\n"), first_missing, last_missing);
	          first_missing = -1;
	       }
	   }

	 /* Do the CRC32 check. Missing sectors are skipped in the CRC report. */

	 if(mode & CHECK_CRC32 && s < eh_sectors)
	 {  guint32 crc = Crc32(buf, 2048); 

	    /* If the CRC buf is exhausted, refill. */

	    if(crcidx >= CRCBUFSIZE)
	    {  size_t remain = ii->sectors-s;
	       size_t size;

	       if(remain < CRCBUFSIZE)
		    size = remain*sizeof(guint32);
	       else size = CRCBUFSIZE*sizeof(guint32);

	       if(fread(crcbuf, 1, size, ecc) != size)
		 Stop(_("Error reading CRC information: %s"),strerror(errno));

	       crcidx = 0;
	    }

            if(crc != crcbuf[crcidx++] && !current_missing)
	    {  PrintLog(_("* CRC error, sector: %lld\n"), s);
	       ii->crcErrors++;
	    }
	 }

	 /* If creation of the CRC32 is requested, do that. */

	 if(mode & CREATE_CRC32)
	 {  crcbuf[crcidx++] = Crc32(buf, 2048);
	    if(crcidx >= CRCBUFSIZE)
	    {  size_t size = CRCBUFSIZE*sizeof(guint32);

	       MD5Update(&ii->md5Ctxt, (char*)crcbuf, size);
	       if(fwrite(crcbuf, 1, size, ecc) != size)
		 Stop(_("Error writing CRC information: %s"),strerror(errno));

	       crcidx = 0;
	    }
	 }

	 MD5Update(&md5ctxt, buf, n);

	 percent = (100*s)/ii->sectors;
	 if(last_percent != percent) 
	 {  PrintMsg(msg,percent);
	    last_percent = percent;
	 }
      }

      /*** Flush the rest of the CRC buffer */

      if((mode & CREATE_CRC32) && crcidx)
      {  size_t size = crcidx*sizeof(guint32);

	 MD5Update(&ii->md5Ctxt, (char*)crcbuf, size);
	 if(fwrite(crcbuf, 1, size, ecc) != size)
	   Stop(_("Error writing CRC information: %s"),strerror(errno));
      }

      /*** The md5sum can only be calculated if all blocks have been successfully read. */

      MD5Final(ii->mediumSum, &md5ctxt);
      PrintMsg(msg,100);
   }

   lseek(ii->file, 0, SEEK_SET);
   g_free(buf);
   return(ii);
}

/*
 * Free the ImageInfo stuff
 */

void FreeImageInfo(ImageInfo *ii)
{
   if(ii->file) close(ii->file);
   g_free(ii);
}

/*
 * Open the .ecc file
 */

static EccInfo* OpenEcc(char *name)
{  EccInfo *ei = g_malloc(sizeof(EccInfo));
   int n;

   ei->name = name;
   ei->eh   = g_malloc(sizeof(EccHeader));

   if(!(ei->file = fopen(name, "r")))
     Stop(_("Can't open %s:\n%s"),name,strerror(errno));
#ifdef SYS_MINGW
   setmode(fileno(ei->file), O_BINARY);
#endif

   n = fread(ei->eh, sizeof(EccHeader), 1, ei->file);
   if(n != 1)
     Stop(_("Can't read ecc header:\n%s"),strerror(errno));

   /*** See if we can use the ecc file */

   if(ei->eh->method_flags[0] != 1 || ei->eh->method_flags[1] || ei->eh->method_flags[2] || ei->eh->method_flags[3])
     Stop(_("\nEcc file has been created with version 0.40.7.\n"
	    "Please use version 0.41.1 or higher to create the .ecc files.\n"
	    "I pledge not to change the file format again until version 2.0 :-)\n"));

   VerifyVersion(ei->eh,1);

   ei->sectors = uchar_to_gint64(ei->eh->sectors);

   return ei;
}

/*
 * Free the EccInfo
 */

void FreeEccInfo(EccInfo *ei)
{
   if(ei->file) fclose(ei->file);
   if(ei->eh) g_free(ei->eh);
   g_free(ei);
}


/*
 * Create ecc information for the medium
 */

static void create_ecc(char *imagename, char *eccname)
{  ImageInfo *ii; 
   FILE *ecc;
   Method *method = FindMethod(Closure->methodName); 

   if(!method) Stop(_("\nMethod %s not available.\n"
		     "Use -m without parameters for a method list.\n"), 
	            Closure->methodName);

   /*** Open the image file and ecc file. */

   if(!(ecc = fopen(eccname, "w")))
     Stop(_("Can't open %s:\n%s"),eccname,strerror(errno));

#ifdef SYS_MINGW
   setmode(fileno(ecc), O_BINARY);
#endif

   ii = OpenImage(imagename, OPEN_FULLSCAN | CREATE_CRC32, ecc);
   PrintMsg("\n");

   if(ii->sectorsMissing)
   {  fclose(ecc);
      unlink(eccname);  /* Do not leave a CRC-only .ecc file behind */
      Stop(_("%lld sectors unread or missing due to errors.\n"),ii->sectorsMissing);
   }

   /*** Create the parity file */

   method->create(ii, eccname, ecc);
   
   /** Clean up */
 
   FreeImageInfo(ii);
   fclose(ecc);
}

/*
 * Fix the medium with ecc information
 */

static void fix(char *imagename, char *eccname)
{  ImageInfo *ii; 
   EccInfo *ei;
   Method *method; 
   char method_name[5];

   /*** Open the image file and header file */

   ii = OpenImage(imagename, OPEN_WRITEABLE, NULL);
   ei = OpenEcc(eccname);

   /*** Fix the image */

   strncpy(method_name, ei->eh->method, 4); method_name[4] = 0;
   method = FindMethod(method_name);
   if(!method) Stop(_("\nUnknown method %s.\n"),method_name);

   method->fix(ii, ei);

   /*** Clean up */

   FreeImageInfo(ii);
   FreeEccInfo(ei);
}

/*
 * Find the place of our executable
 * (Windows only)
 */

#ifdef SYS_MINGW
#ifdef WITH_NLS_YES
static char* get_exe_path()
{  DWORD pid = GetCurrentProcessId();
   HANDLE handle;
   PROCESSENTRY32 pe32;
   HINSTANCE lib;
   DWORD (*GetModuleFileNameExA)(HANDLE,HMODULE,LPSTR,DWORD);

   /* The psapi functions are only available under Win2000 and later. */

   if((lib = LoadLibrary("PSAPI.DLL")))
   {  if((GetModuleFileNameExA = (DWORD(*)(HANDLE,HMODULE,LPSTR,DWORD))GetProcAddress(lib, "GetModuleFileNameExA")))
      {  handle = OpenProcess(PROCESS_QUERY_INFORMATION | PROCESS_VM_READ, 0, pid);
         if(handle)
         {  char path[MAX_PATH];
            int n = GetModuleFileNameExA(handle, 0, path, MAX_PATH);
	    
            CloseHandle(handle);
	    FreeLibrary(lib);

	    if(n>0)
            {  char *backslash = strrchr(path, '\\');

               if(backslash) *backslash=0;
	       return g_strdup(path);
            } 
	    else return NULL;
         }
       }
   }

   /* The following only works for Windows 98 */
       
   handle = CreateToolhelp32Snapshot(TH32CS_SNAPALL, 0);
   if(handle)
   {  pe32.dwSize = sizeof(PROCESSENTRY32);
      if(Process32First(handle, &pe32))
      {  do
         {  if(pe32.th32ProcessID == pid)	
	    {  char *backslash,path[MAX_PATH];
	        
               strcpy(path, pe32.szExeFile);
               backslash = strrchr(path, '\\');
               if(backslash) *backslash=0;

	       return strdup(path);
            }
	 } while(Process32Next(handle, &pe32));
      }
      CloseHandle(handle);
   }

   return NULL;
}
#endif
#endif

/*
 * The all-famous main() loop 
 */

typedef enum
{  MODE_NONE, MODE_CREATE, MODE_READ, MODE_SCAN, MODE_FIX,
   MODE_PRINT, MODE_UNLINK,
   MODE_BYTESET, MODE_ERASE, MODE_RANDOM_ERR, MODE_DRYRUN, MODE_SEQUENCE, MODE_LIST_ASPI,
   MODIFIER_CACHE_SIZE, MODIFIER_PLOT, MODIFIER_DAO, MODIFIER_CLV_SPEED, MODIFIER_CAV_SPEED,
} run_mode;

int main(int argc, char *argv[])
{  static int debug_mode = 0;
   int mode = MODE_NONE; 
   int sequence = MODE_NONE;
   char *device;
   char *imagename = "medium.img";
   char *eccname = "medium.ecc";
   char *debug_arg = NULL, *read_arg = NULL;
#ifdef WITH_NLS_YES
   char *locale_test,src_locale_path[strlen(SRCDIR)+10];
#endif

    /*** Setup the global closure. */

    InitClosure();

    /*** Setup the locale.
         Try the local source directory first (it may be more recent),
         then fall back to the global installation directory.
         The test phrase is supposed to be translated with "yes",
         _independent_ of the actual locale! */

#ifdef WITH_NLS_YES
#ifdef SYS_MINGW
    if(!g_getenv("LANG"))  /* Unix style setting has precedence */
    {  LANGID lang_id;

       /* Try to get locale from Windows
	  and set the respective environment variables. */

       lang_id = GetUserDefaultLangID(); 
       switch(PRIMARYLANGID(lang_id))
       {  case 0x07:  /* German */
	     g_setenv("LANG", "de_DE", 1);
//	     g_setenv("OUTPUT_CHARSET", "iso-8859-1", 1);
	     g_setenv("OUTPUT_CHARSET", "CP850", 1);
	     break;

          case 0x10:  /* Italian */
	     g_setenv("LANG", "it_IT", 1);
//	     g_setenv("OUTPUT_CHARSET", "iso-8859-1", 1);
	     g_setenv("OUTPUT_CHARSET", "CP850", 1);
	     break;

       }
    }
#endif /* SYS_MINGW */

    setlocale(LC_CTYPE, "");
    setlocale(LC_MESSAGES, "");
    textdomain("dvdisaster");

    /* Try local source directory first */

    g_sprintf(src_locale_path,"%s/locale",SRCDIR);
    bindtextdomain("dvdisaster", src_locale_path);

    /* TRANSLATORS: 
       This is a dummy entry which is supposed to translate into "ok".
       Please do not return anything else here. */
    locale_test = gettext("test phrase for verifying the locale installation");

    /* Try current directory. Might work if we are starting
       from the root dir of a binary distribution. */

    if(strcmp(locale_test, "ok"))
    {  char buf[256];

       if(getcwd(buf, 256))
       {  char locale_path[strlen(buf)+20];
          g_sprintf(locale_path,"%s/locale", buf);
          bindtextdomain("dvdisaster", locale_path);
	  locale_test = gettext("test phrase for verifying the locale installation");
       }
    }

#ifdef SYS_MINGW
    /* Try the directory where our executable comes from.
       Only possible under Windows. */

    if(strcmp(locale_test, "ok"))
    {  char *path = get_exe_path();
       char locale_path[strlen(path)+20];

       g_sprintf(locale_path,"%s/locale", path);
       bindtextdomain("dvdisaster", locale_path);
       locale_test = gettext("test phrase for verifying the locale installation");
       g_free(path);
    }
#endif

    /* Last resort: fall back to global locale */

    if(strcmp(locale_test, "ok"))
      bindtextdomain("dvdisaster", LOCALEDIR);  
#endif /* WITH_NLS_YES */

   /*** Greetings first */

#ifndef QUIET
   PrintMsg(_("dvdisaster-%s  Copyright 2004,2005 Carsten Gnoerlich.\n"), VERSION);
   /* TRANSLATORS: Excluding all kinds of warranty might be harmful under your
      legislature. If in doubt, just translate the following like "This is free
      software; please refer to the conditions of the GNU GENERAL PUBLIC LICENSE
      in the source code." Avoid making any legal statements by your own.*/
   PrintMsg(_("This software comes with  ABSOLUTELY NO WARRANTY.  This\n"
	      "is free software and you are welcome to redistribute it\n"
	      "under the conditions of the GNU GENERAL PUBLIC LICENSE.\n"  
	      "See the file \"COPYING\" for further information.\n"));
#endif

   /*** Collect the Method list */

   CollectMethods();

   /*** The following test may go wrong if the compiler chooses the
	wrong packing. */

   if(sizeof(EccHeader) != 4096)
     Stop(_("sizeof(EccHeader) is %d, but must be 4096.\n"),sizeof(EccHeader));

   /*** Determine the default device (OS dependent!) */

   device = DefaultDevice();

   /*** Parse the options */
   
   for(;;)
   {  int option_index,c;
      static struct option long_options[] =
      { {"byteset", 1, 0, MODE_BYTESET },
	{"cache-size", 1, 0, MODIFIER_CACHE_SIZE },
	{"cav", 1, 0, MODIFIER_CAV_SPEED },
	{"clv", 1, 0, MODIFIER_CLV_SPEED },
	{"create", 0, 0, 'c'},
	{"dao", 0, 0, MODIFIER_DAO },
	{"debug", 0, &debug_mode, 1}, 
	{"device", 0, 0, 'd'},
	{"dryrun", 0, 0, MODE_DRYRUN },
        {"ecc", 1, 0, 'e'},
	{"erase", 1, 0, MODE_ERASE },
	{"fix", 0, 0, 'f'},
	{"help", 0, 0, 'h'},
        {"image", 1, 0, 'i'},
	{"jump", 1, 0, 'j'},
#ifdef SYS_MINGW
	{"list", 0, 0, 'l' },
#endif
	{"method", 1, 0, 'm' },
        {"prefix", 1, 0, 'p'},
	{"random-errors", 0, 0, MODE_RANDOM_ERR },
	{"read", 2, 0,'r'},
	{"redundancy", 1, 0, 'n'},
	{"scan", 2, 0,'s'},
	{"test", 0, 0, 't'},
	{"unlink", 0, 0, 'u'},
	{"version", 0, 0, 'v'},
        {0, 0, 0, 0}
      };

      c = getopt_long(argc, argv, 
		      "cd:e:fhi:j:lm::n:p:r::s::tuv",
		      long_options, &option_index);

      if(c == -1) break;

      switch(c)
      {  case 'r': mode = MODE_SEQUENCE; sequence |= 1<<MODE_READ; if(optarg) read_arg = g_strdup(optarg); break;
         case 's': mode = MODE_SEQUENCE; sequence |= 1<<MODE_SCAN; if(optarg) read_arg = g_strdup(optarg); break;
         case 'c': mode = MODE_SEQUENCE; sequence |= 1<<MODE_CREATE; break;
         case 'f': mode = MODE_SEQUENCE; sequence |= 1<<MODE_FIX; break;
         case 't': mode = MODE_SEQUENCE; sequence |= 1<<MODE_PRINT; break;
         case 'u': mode = MODE_SEQUENCE; sequence |= 1<<MODE_UNLINK; break;
         case 'h': mode = MODE_NONE; break;
         case 'i': if(optarg) imagename = g_strdup(optarg); break;
         case 'j': if(optarg) Closure->sectorSkip = atoi(optarg) & ~0xf;
	           if(Closure->sectorSkip<0) Closure->sectorSkip=0;
		   break;
	 case 'l': mode = MODE_LIST_ASPI; break;
         case 'm': if(optarg) 
	           {  g_free(Closure->methodName);
	              Closure->methodName = g_strdup(optarg); 
                   }
	           else {  ListMethods(); FreeClosure(); exit(0); }
	           break;
         case 'n': if(optarg) Closure->redundancy = g_strdup(optarg); break;
         case 'e': if(optarg) eccname = g_strdup(optarg); break;
         case 'p': if(optarg) 
	           {  eccname = g_malloc(strlen(optarg)+5);
		      imagename = g_malloc(strlen(optarg)+5);
		      g_sprintf(eccname,"%s.ecc",optarg);
		      g_sprintf(imagename,"%s.img",optarg);
		   }
	           break;
         case 'd': if(optarg) device = g_strdup(optarg); break;

         case 'v': PrintMsg(_("\ndvdisaster version %s\n\n"), VERSION);
	 	   FreeClosure();
	           exit(EXIT_SUCCESS); 
	           break;

         case  0 : break; /* flag argument */
         case MODIFIER_CACHE_SIZE:
	   Closure->cacheMB = atoi(optarg);
	   if(Closure->cacheMB <   1) 
	     Stop(_("--cache-size must at least be 1MB; 16MB or higher is recommended.")); 
	   if(Closure->cacheMB > 512) 
	     Stop(_("--cache-size maximum is 512MB.")); 
	   break;
         case MODIFIER_DAO: 
	   Closure->noTruncate = 1; 
	   break;
         case MODIFIER_CLV_SPEED:
	   Closure->driveSpeed = atoi(optarg);
	   break;
         case MODIFIER_CAV_SPEED:
	   Closure->driveSpeed = -atoi(optarg);
	   break;
         case MODE_ERASE: 
	   mode = MODE_ERASE;
	   debug_arg = g_strdup(optarg);
	   break;
         case MODE_RANDOM_ERR:
	   mode = MODE_RANDOM_ERR;
	   break;
         case MODE_BYTESET:
	   mode = MODE_BYTESET;
	   debug_arg = g_strdup(optarg);
	   break;
         case MODE_DRYRUN:
	   mode = MODE_DRYRUN;
	   break;
         case '?': mode = MODE_NONE; break;
         default: PrintMsg(_("?? illegal getopt return value %d\n"),c); break;
      }
   }

   /*** Don't allow debugging option if --debug wasn't given */
   
   if(!debug_mode)
     switch(mode)
     {  case MODE_BYTESET:
        case MODE_DRYRUN:
        case MODE_ERASE:
        case MODE_RANDOM_ERR:
	  mode = MODE_NONE;
          break;
     }
	  
   prepare_dead_sector();

   /*** Dispatch action depending on mode.
        The major modes can be executed in sequence 
	(although not all combinations may be really useful) */

   switch(mode)
   {  case MODE_SEQUENCE:
	if(sequence & 1<<MODE_SCAN)
	  ReadMedium(device, imagename, read_arg, 1);

	if(sequence & 1<<MODE_READ)
	  ReadMedium(device, imagename, read_arg, 0);

	if(sequence & 1<<MODE_CREATE)
	  create_ecc(imagename, eccname);

	if(sequence & 1<<MODE_FIX)
	  fix(imagename, eccname);

	if(sequence & 1<<MODE_PRINT)
	  Print(imagename, eccname);

	if(sequence & 1<<MODE_UNLINK)
	{  if(sequence & 1<<MODE_FIX)
	      Stop(_("Deleting the fixed image would lower my karma.\n"));
	   if(unlink(imagename) == 0)
	        PrintMsg(_("\nImage file %s deleted.\n"),imagename);
	   else PrintMsg(_("\nImage file %s not deleted: %s\n"),
			  imagename, strerror(errno));
	}
	 break;

      case MODE_BYTESET:
         Byteset(imagename, debug_arg);
	 break;

      case MODE_DRYRUN:
	 Dryrun(imagename);
	 break;

      case MODE_ERASE:
         Erase(imagename, debug_arg);
	 break;

      case MODE_RANDOM_ERR:
	RandomError(imagename);
	 break;
#ifdef SYS_MINGW
      case MODE_LIST_ASPI:
	 ListAspiDrives();
	 break;
#endif
      default:
	break;
   }

   /*** If no mode was selected, print the help screen. */

   if(mode == MODE_NONE)
   {  
#ifndef QUIET
     /* TRANSLATORS: Program options like -r and --read are not to be translated
	to avoid confusion when discussing the program in international forums. */
      PrintMsg(_("\nCommon usage examples:\n"
	     "  dvdisaster -r,--read   # Read the medium image to hard disc.\n"
	     "                         # Use -rn-m to read a certain sector range, e.g. -r100-200\n"
	     "  dvdisaster -c,--create # Create .ecc information for the medium image.\n"
	     "  dvdisaster -f,--fix    # Try to fix medium image using .ecc information.\n"
	     "  dvdisaster -s,--scan   # Scan the medium for read errors.\n"
	     "  dvdisaster -t,--test   # Test integrity of the .img and .ecc files.\n"
	     "  dvdisaster -u,--unlink # Delete .img files (when other actions complete)\n\n"));

      PrintMsg(_("Drive and file specification:\n"
	     "  -d,--device device     - read from given device   (default: %s)\n"
	     "  -p,--prefix prefix     - prefix of .img/.ecc file (default: medium.*  )\n"
	     "  -i,--image  imagefile  - name of image file       (default: medium.img)\n"
	     "  -e,--ecc    eccfile    - name of parity file      (default: medium.ecc)\n"),
	     device);

#ifdef SYS_MINGW
      PrintMsg(_("  -l,--list              - list drives available under ASPI manager\n\n"));
#else
      PrintMsg("\n");
#endif

      PrintMsg(_("Tweaking options (see manual before using!)\n"
	     "  -j,--jump n            - jump n sectors forward after a read error (default: 16)\n"
	     "  -n,--redundancy n%%     - sets redundancy for error correction (default: 14.3%%)\n"
	     "  --dao                  - assume DAO disc; do not trim image end\n"
	     "  --cache-size n         - image cache size in MB during -c mode (default: 32MB)\n\n"));

      if(debug_mode)
	PrintMsg(_("Debugging options (purposefully undocumented and possibly harmful)\n"
	     "  --debug           - enables the following options\n"
	     "  --byteset s,i,b   - set byte i in sector s to b\n"
	     "  --dryrun          - judge correctability by counting errors/ecc block\n" 
	     "  --erase sector    - erase the given sector\n"
	     "  --erase n-m       - erase sectors n - m, inclusively\n"
	     "  --random-errors   - seed image with (correctable) random errors\n\n"));
#endif
      FreeClosure();
      exit(EXIT_FAILURE);
   }

   FreeClosure();
   exit(EXIT_SUCCESS);
}
