/*  dvdisaster: Additional error correction for optical media.
 *  Copyright (C) 2004,2005 Carsten Gnoerlich.
 *  Project home page: http://www.dvdisaster.com
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA,
 *  or direct your browser at http://www.gnu.org.
 */

#ifndef DVDISASTER_H
#define DVDISASTER_H

/***
 *** I'm too lazy to mess with #include dependencies.
 *** Everything #includeable is rolled up herein...
 */

#define _GNU_SOURCE

#include <glib.h>
#include <glib/gprintf.h>

#include <ctype.h>
#include <errno.h>
#include <fcntl.h>
#include <getopt.h>
#ifdef WITH_NLS_YES
 #include <libintl.h>
 #include <locale.h>
#endif
#include <math.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <stdarg.h>
#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#ifdef SYS_MINGW
#include <io.h>
#endif

#include "md5.h"
#include "version.h"

/* phrase extraction for gettext() */

#ifdef WITH_NLS_YES
 #define _(string) sgettext(string)
#else
 #define _(string) string
#endif

/* Oldest version of dvdisaster which is compatible with us */

#define NEEDED_VERSION 4100

/***
 *** Our global closure (encapsulation of global variables)
 ***/

typedef struct _GlobalClosure
{  int version;         /* Integer number representing current program version */
   char *methodName;    /* Name of currently selected codec */
   int cacheMB;         /* Cache setting for the parity codec, in megabytes */
   int sectorSkip;      /* Number of sectors to skip after read error occurs */
   char *redundancy;    /* Error correction code redundancy */
   int noTruncate;      /* do not truncate image at the end */
   int driveSpeed;      /* currently unused */

   char *deadSector;    /* Copy of our "dead sector" marker. */

} GlobalClosure;

GlobalClosure *Closure;  /* this should be the only global variable! */

/***
 *** An info package about a medium image 
 *** (NOT part or a header of the image file!)
 */

typedef struct _ImageInfo
{  int file;                         /* file handle  for image */
   gint64 size;                      /* number of medium bytes */
   gint64 sectors;                   /* number of medium sectors */
   gint64 sectorsMissing;            /* number of missing medium sectors */
   gint64 crcErrors;                 /* sectors with CRC32 errors */
   int inLast;                       /* bytes in last sector */
   char *name;                       /* name of image file */
   unsigned char mediumFP[16];       /* md5sum of first sector */
   unsigned char mediumSum[16];      /* complete md5sum of whole medium */
   struct MD5Context md5Ctxt;        /* md5 context of crc portion of file */
} ImageInfo;

/***
 *** An info package about a error correction file 
 *** (NOT part or a header of the ecc file!)
 */

typedef struct _EccInfo
{  FILE *file;                  /* file handle for ecc file */
   char *name;                  /* name of ecc file */
   struct _EccHeader *eh;       /* the header struct below */
   gint64 sectors;              /* gint64 version of eh->sectors */ 
} EccInfo;

/***
 *** The .ecc file header
 ***/

typedef struct _EccHeader
{  gint8 cookie[12];           /* "*dvdisaster*" */
   gint8 method[4];            /* e.g. "RS01" */
   gint8 method_flags[4];      /* for free use by the respective methods */
   guint8 mediumFP[16];        /* fingerprint of first medium sector */ 
   guint8 mediumSum[16];       /* complete md5sum of whole medium */
   guint8 eccSum[16];          /* md5sum of ecc code section of .ecc file */
   guint8 sectors[8];          /* number of sectors medium is supposed to have */
   gint32 dataBytes;           /* data bytes per ecc block */
   gint32 eccBytes;            /* ecc bytes per ecc block */
   gint32 creatorVersion;      /* which dvdisaster version created this */
   gint32 neededVersion;       /* oldest version which can decode this file */
   gint32 fpSector;            /* sector used to calculate mediumFP */
   gint8 padding[4000];        /* pad to 4096 bytes: room for future expansion */
} EccHeader;

/***
 *** dvdisaster.c
 ***/

#define OPEN_WRITEABLE    (1<<0)
#define OPEN_FULLSCAN     (1<<1)
#define VERBOSE_FULLSCAN  (1<<2)
#define CREATE_CRC32      (1<<3)
#define CHECK_CRC32       (1<<4)

void CalculateFootprint(ImageInfo*, EccHeader*);
void FreeImageInfo(ImageInfo*);
ImageInfo* OpenImage(char*, int, FILE*);
int VerifyVersion(EccHeader*, int fatal);

/***
 *** closure.c
 ***/

void InitClosure(void);
void FreeClosure(void);

/***
 *** crc32.c
 ***/

guint32 Crc32(unsigned char*, int);

/***
 *** debug.c
 ***/

void Byteset(char*, char*);
void Dryrun(char*);
void Erase(char*, char*);
void RandomError(char*);

/***
 *** method.c / method-link.c
 ***
 * method-link.c is automatically created by the configure script.
 */

typedef struct _Method
{  char name[4];
   char *description;
   void (*create)(ImageInfo*, char*, FILE*);
   void (*dryrun)(ImageInfo*);
   void (*fix)(ImageInfo*, EccInfo*);
} Method;

void BindMethods(void);        /* created by configure in method-link.c */

void CollectMethods(void);
void RegisterMethod(Method*);
void ListMethods(void);
Method* FindMethod(char*);

/***
 *** misc.c 
 ***/

#ifdef SYS_MINGW
#define stat _stati64
#define lseek _lseeki64

int fseeko(FILE*, gint64, int);
int ftruncate(int, gint64);

#define random rand
#define srandom(seed) srand(seed)
#endif

char* sgettext(char*);

gint64 uchar_to_gint64(unsigned char*);
void gint64_to_uchar(unsigned char*, gint64);

void PrintMsg(char*, ...);
void PrintLog(char*, ...);
void Stop(char*, ...);

void AsciiDigest(char*, unsigned char*);
void CalcSectors(gint64, gint64*, int*);

/***
 *** print.c
 ***/

void Print(char*, char*);

/***
 *** print_sense.c
 ***/

void RememberSense(int, int, int);
void PrintSense(char*,...);
void PrintSenseNow(int, int, int);

/*** 
 *** read.c
 ***/

char* DefaultDevice(void);
ImageInfo* ReadMedium(char*, char*, char*, int);

#ifdef SYS_MINGW
void ListAspiDrives(void);
#endif

#endif				/* DVDISASTER_H */
