/*  dvdisaster: Additional error correction for optical media.
 *  Copyright (C) 2004,2005 Carsten Gnoerlich.
 *  Project home page: http://www.dvdisaster.com
 *
 *  The Reed-Solomon error correction draws a lot of inspiration - and even code -
 *  from Phil Karn's excellent Reed-Solomon library: http://www.ka9q.net/code/fec/
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA,
 *  or direct your browser at http://www.gnu.org.
 */

#include "dvdisaster.h"

/***
 *** Method registration
 ***/

static void create(ImageInfo*, char*, FILE*);
static void dryrun(ImageInfo*);
static void fix(ImageInfo*, EccInfo*);

void register_rs01(void)
{  Method *method = g_malloc(sizeof(Method));

   strncpy(method->name, "RS01", 4);
   method->description = _("Classic Reed-Solomon method based on polynomial arithmetic"); 
   method->create = create;
   method->dryrun = dryrun;
   method->fix    = fix;

   RegisterMethod(method);
}

/***
 *** Galois field arithmetic.
 *** 
 * Calculations are done over the extension field GF(2**n).
 * Be careful not to overgeneralize these arithmetics;
 * they only work for the case of GF(p**n) with p being prime.
 */


#define SYMBOLSIZE 8
#define FIELDSIZE (1<<SYMBOLSIZE)
#define FIELDMAX (FIELDSIZE-1)
#define ALPHA0 FIELDMAX

#define GENERATOR_POLY 0x187    /* 1 + X + X**2 + X**7 + X**8 */
#define FIRST_ROOT 112          /* same choices as in CCSDS */                 
#define PRIM_ELEM 11                  
#define PRIMTH_ROOT 116         /* prim-th root of 1 */           

#define	min(a,b) ((a) < (b) ? (a) : (b))

static inline int mod_fieldmax(int x)
{
  while (x >= FIELDMAX) 
  {
    x -= FIELDMAX;
    x = (x >> SYMBOLSIZE) + (x & FIELDMAX);
  }

  return x;
}

/*
 * Interpret our redundancy settings
 *
 * a) "normal" -> nroots=32; "high" -> nroots=64
 * b) "n"      -> nroots = n
 * c) "n%"     -> use closest nroots for given redundancy in percent
 * d) "nm"     -> choose redundancy so that .ecc file does not exceed n megabytes
 */

gint64 ecc_file_size(gint64 sectors, int nr)
{  int nd = FIELDMAX - nr;
   gint64 bytesize; 

   bytesize = 4096 + 4*sectors + 2048*nr*((sectors+nd-1)/nd);

   return (bytesize+0xfffff)/0x100000;   /* size in MB */
}		 

int calculate_redundancy(ImageInfo *ii)
{  int nr = 0;
   char last = 0; 
   double p;
   gint64 fs;

   if(Closure->redundancy) /* get last char of redundancy parameter */
   {  int len = strlen(Closure->redundancy);

      if(len) last = Closure->redundancy[len-1];
   }

   switch(last)
   {  case '%' : p = atof(Closure->redundancy);
                 if(p<3.2 || p>64.5) Stop(_("Redundancy %4.1f%% out of useful range [3.2%%..64.5%%]"),p);
		 nr = (int)round((FIELDMAX*p) / (100.0+p));
	         break;
      case 'm' : fs = strtoll(Closure->redundancy, NULL, 10);
	         if(fs < ecc_file_size(ii->sectors, 8) || fs > ecc_file_size(ii->sectors, 100))
		   Stop(_("Ecc file size %lldm out of useful range [%lld .. %lld]"),
			fs, ecc_file_size(ii->sectors, 8), ecc_file_size(ii->sectors, 100));
		 for(nr=100; nr>8; nr--)
		   if(fs >= ecc_file_size(ii->sectors, nr))
		     break;
	         break;
      default:
	if(!Closure->redundancy || !strcmp(Closure->redundancy, "normal")) nr = 32; 
	else if(!strcmp(Closure->redundancy, "high")) nr = 64;
	else nr = atoi(Closure->redundancy);
	break;
   }

   if(nr < 8 || nr > 100)
     Stop(_("Redundancy %d out of useful range [8..100]."),nr);

   return nr;
}

/* Initialize the Galois field tables */

static int nroots;        /* degree of generator polynomial */
static int ndata;         /* data bytes per ecc block */

static int *gf_index_of;  /* log */
static int *gf_alpha_to;  /* inverse log */
static int *enc_alpha_to; /* inverse log for encoder */
static int *gf_gpoly;     /* generator polynomial */

static void init_tables(int nroots_in)
{  int i,j,b,log,root;

   nroots = nroots_in;
   ndata  = FIELDMAX - nroots;

   /* Allocate the tables.
      The encoder uses a special version of alpha_to which has the mod_fieldmax()
      folded into the table. */

   gf_index_of  = g_malloc(FIELDSIZE * sizeof(int));
   gf_alpha_to  = g_malloc(FIELDSIZE * sizeof(int));
   enc_alpha_to = g_malloc(2*FIELDSIZE * sizeof(int));
   gf_gpoly     = g_malloc((nroots+1) * sizeof(int));
   
   /* create the log/ilog values */

   for(b=1, log=0; log<FIELDMAX; log++)
   {  gf_index_of[b]   = log;
      gf_alpha_to[log] = b;
      b = b << 1;
      if(b & FIELDSIZE)
	b = b ^ GENERATOR_POLY;
   }

   if(b!=1) Stop(_("Failed to create the log tables!\n"));

   /* we're even closed using infinity (makes things easier) */

   gf_index_of[0] = ALPHA0;    /* log(0) = inf */
   gf_alpha_to[ALPHA0] = 0;   /* and the other way around */

   for(b=0; b<2*FIELDSIZE; b++)
     enc_alpha_to[b] = gf_alpha_to[mod_fieldmax(b)];

   /* Create the RS code generator polynomial */

   gf_gpoly[0] = 1;

   for(i=0, root=FIRST_ROOT*PRIM_ELEM; i<nroots; i++, root+=PRIM_ELEM)
   {  gf_gpoly[i+1] = 1;

     /* Multiply gf_gpoly  by  alpha**(root+x) */

     for(j=i; j>0; j--)
     {
       if(gf_gpoly[j] != 0)
         gf_gpoly[j] = gf_gpoly[j-1] ^ gf_alpha_to[mod_fieldmax(gf_index_of[gf_gpoly[j]] + root)];
       else
	 gf_gpoly[j] = gf_gpoly[j-1];
     }

     gf_gpoly[0] = gf_alpha_to[mod_fieldmax(gf_index_of[gf_gpoly[0]] + root)];
   }

   /* Store the polynomials index for faster encoding */ 

   for(i=0; i<=nroots; i++)
     gf_gpoly[i] = gf_index_of[gf_gpoly[i]];

#if 0
   /* for the precalculated unrolled loops only */

   for(i=nroots-1; i>0; i--)
     PrintMsg(
	    "                  par_idx[((++spk)&31)] ^= enc_alpha_to[feedback + %3d];\n",
	    gf_gpoly[i]);

   PrintMsg("                  par_idx[sp] = enc_alpha_to[feedback + %3d];\n",
	  gf_gpoly[0]);
#endif

}

static void free_tables(void)
{
  g_free(gf_index_of);
  g_free(gf_alpha_to);
  g_free(enc_alpha_to);
  g_free(gf_gpoly);
}

/*
 * Special debugging options which are only useful during
 * development and back testing. Having them on in production
 * releases is not recommended.
 */

/*
 * Read a medium sector from the .img file.
 *
 * Two special cases here:
 * - Missing sectors (beyond the range recorded in eh->sectors) will be padded with zeros,
 *   since we need a multiple of ndata sectors for the parity generation. 
 * - Missing sectors beyond the range recorded in ii->sectors, but before the real end
 *   as defined above are treated as "dead sectors".
 */

//#define DEBUG_ECC
//#define DEBUG_READ

#ifdef DEBUG_READ
char bucket[3000000];
int deadcnt;
int padcnt;
int seccnt;
#endif

static void read_sector(ImageInfo *ii, EccHeader *eh, char *buf, gint64 s)
{ gint64 eh_sectors = uchar_to_gint64(eh->sectors);

  if(s >= ii->sectors && s < eh_sectors)
  {
     memcpy(buf, Closure->deadSector, 2048);   /* truncated image / dead sector */
#ifdef DEBUG_READ
     deadcnt++;
#endif
  }
  else if(s >= eh_sectors)      
  {
     memset(buf, 0, 2048);            /* zero padding for reads past the image */
#ifdef DEBUG_READ
     padcnt++;
#endif
  }
  else                                /* else normal read within the image */
  {  int n,expected;
	
     if(lseek(ii->file, (gint64)(2048*s), SEEK_SET) == -1)
       Stop(_("lseek failed for %s, sector %lld"),ii->name,s);

     /* Prepare for short reads at the last image sector.
	Doesn't happen for CD and DVD media, but perhaps for future media? */

     if(s < ii->sectors-1) expected = 2048;
     else  
     {  memset(buf, 0, 2048);
        expected = ii->inLast;
     }

     /* Finally, read the sector */

     n = read(ii->file, buf, expected);
     if(n != expected)
       Stop(_("could not read medium sector %lld:\n%s"),s,strerror(errno));

#ifdef DEBUG_READ
     seccnt++;
     bucket[s]++;
#endif
  }
}

/*
 * Read crc values from the .ecc file.
 */

static void read_crc(FILE *ecc, guint32 *buf, int first_sector, int n_sectors)
{  int n;
  
   if(fseeko(ecc, (gint64)(sizeof(EccHeader) + first_sector*sizeof(guint32)), SEEK_SET) == -1)
     Stop(_("fseek failed reading crc"));

   n = fread(buf, sizeof(guint32), n_sectors, ecc);
	
   if(n != n_sectors)
     Stop(_("problem reading crc data: %s"),strerror(errno));
}

/***
 *** Fix the medium sectors.
 ***
 * Note that if DEBUG_ECC is defined, we check the syndromes of each ecc block
 * regardless whether its CRC sum is correct (slow!), but do not actually
 * correct anything in the image file. 
 * This mode is just for testing the integrity of the RS codec and works
 * best with a complete/undamaged image file.
 */

static void fix(ImageInfo *ii, EccInfo *ei)
{  EccHeader *eh = ei->eh;
   unsigned char *img_block[256];
   unsigned char parity[256];
   int erasure_count,erasure_list[256],erasure_map[256];
   guint32 *crc_buf[256];
   int unexpected_failure;
   gint64 block_idx[256];
   gint64 s,si;
   int i,j,k,n;
   gint64 corrected, uncorrected;
   gint64 parity_block = 0;
   int worst_ecc,damaged_ecc,damaged_sec,percent,last_percent = -1;
   int cache_size,cache_sector,cache_offset = 0;

   /*** Announce what we are going to do. */

#ifdef DEBUG_ECC
   PrintMsg(_("\n***\n*** Fix mode with DEBUG_ECC settings: Image will NOT be repaired!\n***\n"));
#else
   PrintMsg(_("\nFix mode: Repairable sectors will be fixed in the image.\n"));
#endif

   /*** Do some trivial comparisons between the .ecc file and the image file */

   if(ii->sectors > ei->sectors)
   { int diff = ii->sectors - ei->sectors;

     if(diff>2) 
       Stop(_("Image file has more sectors than ecc file (%lld > %lld)\n"
	      "This is impossible.\n"), ii->sectors, ei->sectors);
     else
     {  PrintMsg(_("* Warning: Image file is %d sectors longer than expected.\n"
	           "*          Assuming this is a TAO mode medium.\n"
	           "*          %d sectors will be removed from the image end.\n"),
		 diff, diff);

        ii->sectors -= diff;

        if(ftruncate(ii->file, (gint64)(2048*ii->sectors)))
	  Stop(_("Could not truncate %s: %s\n"),ii->name,strerror(errno));
     }
   }

   if(ii->sectors < ei->sectors)
     PrintMsg(_("* Warning: Image file appears to be truncated.\n"
	        "*          Consider completing it with another -r pass before going on.\n"));

   /* Compatibility with version 0.60 and beyond */

   CalculateFootprint(ii, eh);

   if(!memcmp(ii->mediumFP, "\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000", 16))
     PrintMsg(_("* Warning: Sector 0 is missing. Can not compare image and ecc fingerprints.\n"
	        "*          Double check that image and ecc file belong together.\n"));
   else if(memcmp(ii->mediumFP, eh->mediumFP, 16))
          Stop(_("Footprints of image and ecc file do not match.\n"
	         "Image and ecc file do not belong together.\n"));

   /*** Set up the Galois field arithmetic */

   init_tables(eh->eccBytes);

   /*** Prepare buffers for ecc code generation.
	We assemble ecc blocks from ndata medium sectors spread over the full medium size,
        and read cache_size * ndata medium sectors ahead. */

   cache_size = 2*Closure->cacheMB;  /* ndata medium sectors are approx. 0.5MB */

   for(i=0; i<ndata; i++)
   {  img_block[i] = g_malloc(cache_size*2048);
      crc_buf[i]   = g_malloc(sizeof(int) * cache_size);
   }

   /*** Setup the block counters for mapping medium sectors to
	ecc blocks */

   s = (ei->sectors+ndata-1)/ndata;

   for(si=0, i=0; i<ndata; si+=s, i++)
     block_idx[i] = si;

   cache_sector = cache_size;  /* forces instant reload of cache */

   /*** Verify ecc information for the medium image. */ 

   corrected = uncorrected = 0;
   worst_ecc = damaged_ecc = damaged_sec = 0;

   for(si=0; si<s; si++)
   { 
     /* Read the next batch of (cache_size * ndata) medium sectors
        if the cache ran empty. */

     if(cache_sector >= cache_size)
     {  
        if(s-si < cache_size)
           cache_size = s-si;
        for(i=0; i<ndata; i++)
        {  int offset = 0;
	   for(j=0; j<cache_size; j++) 
	   {  read_sector(ii, eh, img_block[i]+offset, block_idx[i]+j);
	      offset += 2048;
	   }
	   read_crc(ei->file, crc_buf[i], block_idx[i], cache_size);
	}
        cache_sector = cache_offset = 0;
     }

     /* Determine erasures based on the "dead sector" marker */

     erasure_count = 0;
     unexpected_failure = 0;

     for(i=0; i<ndata; i++)
     {  guint32 crc = Crc32(img_block[i]+cache_offset, 2048);

        erasure_map[i] = 0;

        if(block_idx[i] < ei->sectors)  /* ignore the padding sectors! */
	{
	  if(!memcmp(img_block[i]+cache_offset, Closure->deadSector, 2048))
	  {  erasure_map[i] = 1;
	     erasure_list[erasure_count++] = i;
          }
	  else if(crc != crc_buf[i][cache_sector])
	  {  erasure_map[i] = 3;
	     erasure_list[erasure_count++] = i;
	     PrintLog(_("CRC error in sector %lld\n"),block_idx[i]);
          }
	}
     }

     if(!erasure_count)  /* Skip completely read blocks */
     {  
#ifndef DEBUG_ECC
        parity_block+=2048;
        goto skip;
#else
	;
#endif
     }
     else
     {  damaged_ecc++;
        damaged_sec+=erasure_count;
     }

     if(erasure_count>worst_ecc)
       worst_ecc = erasure_count;

     /* Turn the ndata medium sectors into 2048 ecc blocks
        and try to correct them. */

     if(erasure_count>nroots)   /* uncorrectable */
     {  PrintLog(_("* %3d unrepairable sectors: "), erasure_count);
        for(i=0; i<erasure_count; i++)
	  PrintLog("%lld ", block_idx[erasure_list[i]]);
	PrintLog("\n");
	uncorrected += erasure_count;
	parity_block+=2048;

#ifndef DEBUG_ECC
	/* For truncated images, make sure we leave no "zero holes" in the image
	   by writing the sector(s) with our "dead sector" markers. */

	for(i=0; i<erasure_count; i++)
	{  gint64 idx = block_idx[erasure_list[i]];

	   if(idx < ii->sectors)
	     continue;  /* It's (already) dead, Jim ;-) */

	   if(lseek(ii->file, (gint64)(2048*idx), SEEK_SET) == -1)
	     Stop(_("lseek/dead sector failed recovering %s"),ii->name);

	   n = write(ii->file, Closure->deadSector, 2048);
	   if(n != 2048)
	     Stop(_("could not write dead sector marker for sector %lld:\n%s"),idx,strerror(errno));
	}
#endif
     }
     else  /* try to correct them */
     {  int bi;

        for(bi=0; bi<2048; bi++)
        {  int offset = cache_offset+bi;
	   int r, deg_lambda, el, deg_omega;
	   int u,q,tmp,num1,num2,den,discr_r;
	   int lambda[nroots+1], s[nroots]; /* Err+Eras Locator poly * and syndrome poly */
	   int b[nroots+1], t[nroots+1], omega[nroots+1];
	   int root[nroots], reg[nroots+1], loc[nroots];
	   int syn_error, count;

	   /* Read the parity bytes */

	   if(fseeko(ei->file, (gint64)(sizeof(EccHeader) + ei->sectors*sizeof(guint32) + nroots*parity_block), SEEK_SET) == -1)
	      Stop(_("fseek/sector failed for %s"), ei->name);

	   n = fread(parity, 1, nroots, ei->file);
	   if(n != nroots)
	     Stop(_("Can't read ecc file:\n%s"),strerror(errno));
	   parity_block++;

	   /* Form the syndromes; i.e., evaluate data(x) at roots of g(x) */

	   for(i=0; i<nroots; i++)
	     s[i] = img_block[0][offset];

	   for(j=1; j<FIELDMAX; j++)
	   {  int data = j>=ndata ? parity[j-ndata] : img_block[j][offset];

	      for(i=0;i<nroots;i++)
              {  if(s[i] == 0) s[i] = data;
	         else s[i] = data ^ gf_alpha_to[mod_fieldmax(gf_index_of[s[i]] + (FIRST_ROOT+i)*PRIM_ELEM)];
	      }
	   }

	   /* Convert syndromes to index form, check for nonzero condition */

	   syn_error = 0;
	   for(i=0; i<nroots; i++)
           {  syn_error |= s[i];
	      s[i] = gf_index_of[s[i]];
	   }

	   /* If it is already correct by coincidence,
	      we have nothing to do any further */

	   if(!syn_error) continue;

	   /* NOTE: Since we already know all our erasure positions, 
	      we could do away simpler than by using the Berlekamp and Chien
	      algorithms.I've left them in this release to have a reference
	      implementation Phil's library code which can be compared
	      against later optimized versions. */

	   /* Init lambda to be the erasure locator polynomial */

	   memset(lambda+1, 0, nroots*sizeof(lambda[0]));
	   lambda[0] = 1;

	   lambda[1] = gf_alpha_to[mod_fieldmax(PRIM_ELEM*(FIELDMAX-1-erasure_list[0]))];
	   for(i=1; i<erasure_count; i++) 
	   {  u = mod_fieldmax(PRIM_ELEM*(FIELDMAX-1-erasure_list[i]));
	      for(j=i+1; j>0; j--) 
	      {  tmp = gf_index_of[lambda[j-1]];
	         if(tmp != ALPHA0)
		   lambda[j] ^= gf_alpha_to[mod_fieldmax(u + tmp)];
	      }
	   }
	
	   for(i=0; i<nroots+1; i++)
	     b[i] = gf_index_of[lambda[i]];
  
	   /* Begin Berlekamp-Massey algorithm to determine error+erasure locator polynomial */

	   r = erasure_count;   /* r is the step number */
	   el = erasure_count;
	   while(++r <= nroots) /* Compute discrepancy at the r-th step in poly-form */
	   {  
	     discr_r = 0;
	     for(i=0; i<r; i++)
	       if((lambda[i] != 0) && (s[r-i-1] != ALPHA0))
		 discr_r ^= gf_alpha_to[mod_fieldmax(gf_index_of[lambda[i]] + s[r-i-1])];

	     discr_r = gf_index_of[discr_r];	/* Index form */

	     if(discr_r == ALPHA0) 
	     {
	        /* B(x) = x*B(x) */
	        memmove(b+1, b, nroots*sizeof(b[0]));
	        b[0] = ALPHA0;
	     } 
	     else 
	     {  /* T(x) = lambda(x) - discr_r*x*b(x) */
	        t[0] = lambda[0];
	        for(i=0; i<nroots; i++) 
		{  if(b[i] != ALPHA0)
		        t[i+1] = lambda[i+1] ^ gf_alpha_to[mod_fieldmax(discr_r + b[i])];
		   else t[i+1] = lambda[i+1];
		}

		if(2*el <= r+erasure_count-1) 
		{  el = r + erasure_count - el;

		   /* B(x) <-- inv(discr_r) * lambda(x) */
		   for(i=0; i<=nroots; i++)
		     b[i] = (lambda[i] == 0) ? ALPHA0 : mod_fieldmax(gf_index_of[lambda[i]] - discr_r + FIELDMAX);
		} 
		else 
		{  /* 2 lines below: B(x) <-- x*B(x) */
		   memmove(b+1, b, nroots*sizeof(b[0]));
		   b[0] = ALPHA0;
		}

		memcpy(lambda,t,(nroots+1)*sizeof(t[0]));
	     }
	   }

	   /* Convert lambda to index form and compute deg(lambda(x)) */
	   deg_lambda = 0;
	   for(i=0; i<nroots+1; i++)
	   {  lambda[i] = gf_index_of[lambda[i]];
	      if(lambda[i] != ALPHA0)
		deg_lambda = i;
	   }

	   /* Find roots of the error+erasure locator polynomial by Chien search */
	   memcpy(reg+1, lambda+1, nroots*sizeof(reg[0]));
	   count = 0;		/* Number of roots of lambda(x) */

	   for(i=1, k=PRIMTH_ROOT-1; i<=FIELDMAX; i++, k=mod_fieldmax(k+PRIMTH_ROOT))
	   {  q=1; /* lambda[0] is always 0 */

	      for(j=deg_lambda; j>0; j--)
	      {  if(reg[j] != ALPHA0) 
		 {  reg[j] = mod_fieldmax(reg[j] + j);
		    q ^= gf_alpha_to[reg[j]];
		 }
	      }

	      if(q != 0) continue; /* Not a root */

	      /* store root (index-form) and error location number */

	      root[count] = i;
	      loc[count] = k;

	      /* If we've already found max possible roots, abort the search to save time */

	      if(++count == deg_lambda) break;
	      //if(++count >= deg_lambda) break;
	   }

	   /* deg(lambda) unequal to number of roots => uncorrectable error detected */

	   if(deg_lambda != count)
	   {  PrintMsg(_("Decoder problem (%d != %d) for %d sectors: "), deg_lambda, count, erasure_count);

	      for(i=0; i<erasure_count; i++)
	      {  gint64 idx = block_idx[erasure_list[i]];
	       
                 PrintLog("%lld ", idx);
	      }
	      PrintLog("\n");
	      break;
	   }

	   /* Compute err+eras evaluator poly omega(x) = s(x)*lambda(x) 
	      (modulo x**nroots). in index form. Also find deg(omega). */

	   deg_omega = deg_lambda-1;

	   for(i=0; i<=deg_omega; i++)
	   {  tmp = 0;
	      for(j=i; j>=0; j--)
	      {  if((s[i - j] != ALPHA0) && (lambda[j] != ALPHA0))
		   tmp ^= gf_alpha_to[mod_fieldmax(s[i - j] + lambda[j])];
	      }

	      omega[i] = gf_index_of[tmp];
	   }

	   /* Compute error values in poly-form. 
	      num1 = omega(inv(X(l))), 
              num2 = inv(X(l))**(FIRST_ROOT-1) and 
	      den  = lambda_pr(inv(X(l))) all in poly-form. */

	   for(j=count-1; j>=0; j--)
	   {  num1 = 0;

	      for(i=deg_omega; i>=0; i--) 
	      {  if(omega[i] != ALPHA0)
		    num1 ^= gf_alpha_to[mod_fieldmax(omega[i] + i * root[j])];
	      }

	      num2 = gf_alpha_to[mod_fieldmax(root[j] * (FIRST_ROOT - 1) + FIELDMAX)];
	      den = 0;
    
	      /* lambda[i+1] for i even is the formal derivative lambda_pr of lambda[i] */

	      for(i=min(deg_lambda, nroots-1) & ~1; i>=0; i-=2) 
	      {  if(lambda[i+1] != ALPHA0)
		   den ^= gf_alpha_to[mod_fieldmax(lambda[i+1] + i * root[j])];
	      }

	      /* Apply error to data */

	      if(num1 != 0)
	      {  int location = loc[j];
		
		 if(location >= 0 && location < ndata)
		 {  if(erasure_map[location] == 3)
		    {  int old = img_block[location][offset];
		       int new = old ^ gf_alpha_to[mod_fieldmax(gf_index_of[num1] + gf_index_of[num2] + FIELDMAX - gf_index_of[den])];

		       PrintLog(_("-> Error located in sector %lld at byte %4d (value %2x '%c', expected %2x '%c')\n"),
			        block_idx[location], bi, 
			        old, isprint(old) ? old : '.',
			        new, isprint(new) ? new : '.');
		    }

		    if(!erasure_map[location])
		      PrintMsg(_("Unexpected byte error in sector %lld, byte %d\n"),
			       block_idx[location], bi);

		    img_block[location][offset] ^= gf_alpha_to[mod_fieldmax(gf_index_of[num1] + gf_index_of[num2] + FIELDMAX - gf_index_of[den])];
		 }
		 else
		   PrintMsg(_("Bad error location %d; corrupted .ecc file?\n"), location);
	      }
	   }
	}
     }

     /*** Report if any sectors could be recovered.
	  Write the recovered sectors to the image file .*/

     if(erasure_count && erasure_count<=nroots)
     {  PrintLog(_("  %3d repaired sectors: "), erasure_count);

        for(i=0; i<erasure_count; i++)
	{  gint64 idx = block_idx[erasure_list[i]];
	   int length;

	   PrintLog("%lld ", idx);

#ifndef DEBUG_ECC
	   /* Write the recovered sector */

	   if(lseek(ii->file, (gint64)(2048*idx), SEEK_SET) == -1)
	     Stop(_("lseek/sector failed recovering %s"),ii->name);

	   if(idx < ei->sectors-1) length = 2048;
	   else length = ii->inLast;  /* error: use inLast calculated from eh->sectors */

	   n = write(ii->file, cache_offset+img_block[erasure_list[i]], length);
	   if(n != length)
	     Stop(_("could not write medium sector %lld:\n%s"),idx,strerror(errno));
#endif
	}
	PrintLog("\n");
	corrected += erasure_count;
     }

skip:
     /* Advance the cache pointers */

     cache_sector++;
     cache_offset += 2048;

     /* Report progress */

     percent = (1000*si)/s;
     if(last_percent != percent) 
     {  PrintMsg(_("Ecc progress: %3d.%1d%%\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b"),
		percent/10,percent%10);
        last_percent = percent;
     }

     /* Increment the block indices */

     for(i=0; i<ndata; i++)
	block_idx[i]++;
   }

   /*** Clean up */

   PrintMsg(_("Ecc progress: 100.0%%\n"));
   if(corrected > 0) PrintMsg(_("Repaired sectors: %lld     \n"),corrected);
   if(uncorrected > 0) PrintMsg(_("Unrepaired sectors: %lld\n"), uncorrected);      
   else
   {  if(!corrected)
           PrintMsg(_("Good! All sectors already present.\n"));
      else PrintMsg(_("Good! All sectors are repaired.\n"));
   }
   if(corrected > 0 || uncorrected > 0)
     PrintMsg(_("Erasure counts per ecc block:  avg =  %.1f; worst = %d.\n"),
	     (double)damaged_sec/(double)damaged_ecc,worst_ecc);

   for(i=0; i<ndata; i++)
   {  g_free(img_block[i]);
      g_free(crc_buf[i]);
   }

   free_tables();

#ifdef DEBUG_READ
   for(i=0; i<(int)ei->sectors; i++)
     if(bucket[i] != 1)
       PrintMsg("sector %d: %d reads\n",i,bucket[i]);

   PrintMsg("Sectors: %d expected, %d dead, %d padded, %d read.\n",(int)ei->sectors ,deadcnt,padcnt,seccnt); 
#endif
}

/***
 *** Dry run: Simply count the damaged medium sectors per ecc block.
 ***
 * This is for debugging purposes only; 
 * it will just serve as a quick assessment of medium correctability
 * without needing an actual .ecc file.
 */

static void dryrun(ImageInfo *ii)
{  EccHeader eh;
   unsigned char *img_block[256];
   int erasure_count,erasure_list[256];
   gint64 block_idx[256];
   gint64 s,si;
   int i,j;
   gint64 corrected, uncorrected;
   int worst_ecc,damaged_ecc,damaged_sec,percent,last_percent = -1;
   int cache_size,cache_sector,cache_offset = 0;

   /*** Calculate max erasure and data counts */

   nroots = calculate_redundancy(ii);
   ndata  = FIELDMAX - nroots;

   /*** Announce what we are going to do. */

   PrintMsg(_("\nDryrun mode: Just estimates repairability by counting damaged sectors\n"
	      "             per parity block (without actually using/requiring the .ecc file)\n"));

   /*** Prepare buffers for ecc code generation.
	We assemble ecc blocks from ndata medium sectors spread over the full medium size,
        and read cache_size * ndata medium sectors ahead. */

   cache_size = 2*Closure->cacheMB;  /* ndata medium sectors are approx. 0.5MB */

   for(i=0; i<ndata; i++)
     img_block[i] = g_malloc(cache_size*2048);

   /*** Setup the block counters for mapping medium sectors to
	ecc blocks */

   gint64_to_uchar(eh.sectors, ii->sectors);
   s = (ii->sectors+ndata-1)/ndata;

   for(si=0, i=0; i<ndata; si+=s, i++)
     block_idx[i] = si;

   cache_sector = cache_size;  /* forces instant reload of cache */

   corrected = uncorrected = 0;
   worst_ecc = damaged_ecc = damaged_sec = 0;

   for(si=0; si<s; si++)
   { 
     /* Read the next batch of (cache_size * ndata) medium sectors
        if the cache ran empty. */

     if(cache_sector >= cache_size)
     {  
        if(s-si < cache_size)
           cache_size = s-si;
        for(i=0; i<ndata; i++)
        {  int offset = 0;
	   for(j=0; j<cache_size; j++) 
	   {  read_sector(ii, &eh, img_block[i]+offset, block_idx[i]+j);
	      offset += 2048;
	   }
	}
        cache_sector = cache_offset = 0;
     }

     /* Determine erasures based on the "dead sector" marker */

     erasure_count = 0;

     for(i=0; i<ndata; i++)
     {  if(!memcmp(img_block[i]+cache_offset, Closure->deadSector, 2048))
	  erasure_list[erasure_count++] = i;
     }

     if(!erasure_count)  /* Skip completely read blocks */
        goto skip;
     else 
     {  damaged_ecc++;
        damaged_sec+=erasure_count;
     }

     if(erasure_count>worst_ecc)
       worst_ecc = erasure_count;

     /* Judge correctablity by counting the number of erasures. */

     if(erasure_count>nroots)
     {  PrintLog(_("* %3d unrepairable sectors: "), erasure_count);
        for(i=0; i<erasure_count; i++)
	  PrintLog("%lld ", block_idx[erasure_list[i]]);
	PrintLog("\n");
	uncorrected += erasure_count;
     }
     
     if(erasure_count && erasure_count<=nroots)
     {  PrintLog(_("  %3d repairable sectors: "), erasure_count);
        for(i=0; i<erasure_count; i++)
	{  PrintLog("%lld ", block_idx[erasure_list[i]]);
	}
	PrintLog("\n");
	corrected += erasure_count;
     }

skip:
     /* Advance the cache pointers */

     cache_sector++;
     cache_offset += 2048;

     /* Report progress */

     percent = (1000*si)/s;
     if(last_percent != percent) 
     {  PrintMsg(_("Progress: %3d.%1d%%\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b"),percent/10,percent%10);
        last_percent = percent;
     }

     /* Increment the block indices */

     for(i=0; i<ndata; i++)
	block_idx[i]++;
   }

   /*** Clean up */

   PrintMsg(_("Progress: 100.0%%\n"));
   if(corrected > 0) PrintMsg(_("Repairable sectors  : %lld\n"),corrected);
   if(uncorrected > 0) PrintMsg(_("Unrepairable sectors: %lld\n"), uncorrected);      
   else
   {  if(!corrected)
           PrintMsg(_("Good! All sectors are already present.\n"));
      else PrintMsg(_("Good! All sectors are repairable.\n"));
   }
   if(corrected > 0 || uncorrected > 0)
     PrintMsg(_("Erasure counts per ecc block:  avg =  %.1f; worst = %d.\n"),
	     (double)damaged_sec/(double)damaged_ecc,worst_ecc);

   for(i=0; i<ndata; i++)
     g_free(img_block[i]);
}

/***
 *** Create parity information for the medium sectors.
 ***/

/*
 * Create the parity file.
 */

enum { NORMAL, HIGH, GENERIC };

static void create(ImageInfo *ii, char *eccname, FILE *ecc)
{  EccHeader eh;
   unsigned char *data;
   unsigned char *parity;
   gint64 block_idx[256];  /* must be >= ndata */
   gint64 s,si;
   int i,n;
   int percent,max_percent,progress = 0, last_percent = -1;
   int n_parity_blocks,n_layer_sectors;
   int n_parity_bytes,n_layer_bytes;
   int layer,chunk;
   int loop_type = GENERIC;

   /*** Set up the Galois field arithmetic */

   /* Calculate number of roots (= max. number of erasures)
      and number of data bytes from redundancy setting */

   if(!Closure->redundancy || !strcmp(Closure->redundancy, "normal")) 
	                                          loop_type = NORMAL;
   else if(!strcmp(Closure->redundancy, "high"))  loop_type = HIGH;

   n = calculate_redundancy(ii);
   init_tables(n);

   PrintMsg(_("Encoding with Method RS01: %d roots, %4.1f%% redundancy.\n"),
	    nroots,((double)nroots*100.0)/(double)ndata);

   /*** Prepare Ecc file header.
        The .eccSum will be filled in after all ecc blocks have been created. */

   memset(&eh,0,sizeof(EccHeader));
   strncpy(eh.cookie, "*dvdisaster*", 12);
   strncpy(eh.method, "RS01", 4);
   eh.method_flags[0] = 1;
   gint64_to_uchar(eh.sectors, ii->sectors);
   eh.dataBytes       = ndata;
   eh.eccBytes        = nroots;

   eh.creatorVersion  = Closure->version;
   eh.neededVersion   = nroots == 32 ? 4100 : 5200;

   memcpy(eh.mediumFP, ii->mediumFP, 16);
   memcpy(eh.mediumSum, ii->mediumSum, 16);

   if(fseeko(ecc, (gint64)sizeof(EccHeader) + ii->sectors*sizeof(guint32), SEEK_SET) == -1)
	Stop(_("fseek failed skipping ecc+crc header: %s"),strerror(errno));

   /*** Allocate buffers for the parity calculation and image data caching. 

        The algorithm builds the parity file consecutively in chunks of n_parity_blocks.
        We use all the amount of memory allowed by cacheMB for caching the parity blocks. */

   n_parity_blocks = (Closure->cacheMB<<20) / nroots;
   n_parity_blocks &= ~0x7ff;                   /* round down to multiple of 2048 */
   n_parity_bytes  = nroots * n_parity_blocks;

   /* Each chunk of parity blocks is built iteratively by processing the data in layers
      (first all bytes at pos 0, then pos 1, until ndata layers have been processed).
      So one buffer of n_layer_bytes = n_parity_blocks needs to be buffered.
      For practical reasons we require that the layer size is a multiple of the
      medium sector size of 2048 bytes. */

   n_layer_bytes   = n_parity_blocks;
   n_layer_sectors = n_parity_blocks/2048;

   if(n_layer_sectors*2048 != n_parity_blocks)
     Stop(_("Internal error: parity blocks are not a multiple of sector size.\n"));

   parity = g_malloc(n_parity_bytes);
   data   = g_malloc(n_layer_bytes);

   /*** Setup the block counters for mapping medium sectors to ecc blocks 
        The image is divided into ndata sections;
        with each section spanning s sectors. */

   s = (ii->sectors+ndata-1)/ndata;

   for(si=0, i=0; i<ndata; si+=s, i++)
     block_idx[i] = si;

#if 0
PrintMsg("Block counters: %lld, %lld, %lld, ..., %lld, %lld\n",
         block_idx[0],block_idx[1],block_idx[2],block_idx[ndata-2],block_idx[ndata-1]);
#endif

   /*** Create ecc information for the medium image. */ 

   max_percent = ndata * ((s / n_layer_sectors) + 1);

   /* Process the image.
      From each section a chunk of n_layer_sectors is read in at once.
      So after (s/n_layer_sectors)+1 iterations the whole image has been processed. */

   for(chunk=0; chunk<s; chunk+=n_layer_sectors) 
   {  int actual_layer_bytes,actual_layer_sectors;

      /* Prepare the parity data for the next chunk. */

      memset(parity, 0, n_parity_bytes);

      /* The last chunk may contain fewer sectors. */

      if(chunk+n_layer_sectors < s)
           actual_layer_sectors = n_layer_sectors;
      else actual_layer_sectors = s-chunk;

      actual_layer_bytes   = 2048*actual_layer_sectors;

      /* Work each of the ndata data layers 
	 into the parity data of the current chunk. */

      switch(loop_type)
      { case NORMAL:  /* Inner loop unrolled for nroots = 32. */
	{int sp=1;    /* sp==1 makes sure sp==0 after ndata bytes [since (223+1) mod 32 = 0]*/
  
         for(layer=0; layer<ndata; layer++)
	 {  int offset = 0;
            unsigned char *par_idx = parity;

	    /* Read the next data sectors of this layer. */

	    for(i=0; i<actual_layer_sectors; i++)
	    {  read_sector(ii, &eh, data+offset, block_idx[layer]);
	       block_idx[layer]++;
	       offset += 2048;
	    }

	    /* Now process the data bytes of the current layer. */

	    for(i=0; i<actual_layer_bytes; i++)
	    {  register int feedback;

	       feedback = gf_index_of[data[i] ^ par_idx[sp]];

	       if(feedback != ALPHA0) /* non-zero feedback term */
	       {  register int spk = sp;

                  par_idx[((++spk)&31)] ^= enc_alpha_to[feedback + 249];
                  par_idx[((++spk)&31)] ^= enc_alpha_to[feedback +  59];
                  par_idx[((++spk)&31)] ^= enc_alpha_to[feedback +  66];
                  par_idx[((++spk)&31)] ^= enc_alpha_to[feedback +   4];
                  par_idx[((++spk)&31)] ^= enc_alpha_to[feedback +  43];
                  par_idx[((++spk)&31)] ^= enc_alpha_to[feedback + 126];
                  par_idx[((++spk)&31)] ^= enc_alpha_to[feedback + 251];
                  par_idx[((++spk)&31)] ^= enc_alpha_to[feedback +  97];
                  par_idx[((++spk)&31)] ^= enc_alpha_to[feedback +  30];
                  par_idx[((++spk)&31)] ^= enc_alpha_to[feedback +   3];
                  par_idx[((++spk)&31)] ^= enc_alpha_to[feedback + 213];
                  par_idx[((++spk)&31)] ^= enc_alpha_to[feedback +  50];
                  par_idx[((++spk)&31)] ^= enc_alpha_to[feedback +  66];
                  par_idx[((++spk)&31)] ^= enc_alpha_to[feedback + 170];
                  par_idx[((++spk)&31)] ^= enc_alpha_to[feedback +   5];
                  par_idx[((++spk)&31)] ^= enc_alpha_to[feedback +  24];
                  par_idx[((++spk)&31)] ^= enc_alpha_to[feedback +   5];
                  par_idx[((++spk)&31)] ^= enc_alpha_to[feedback + 170];
                  par_idx[((++spk)&31)] ^= enc_alpha_to[feedback +  66];
                  par_idx[((++spk)&31)] ^= enc_alpha_to[feedback +  50];
                  par_idx[((++spk)&31)] ^= enc_alpha_to[feedback + 213];
                  par_idx[((++spk)&31)] ^= enc_alpha_to[feedback +   3];
                  par_idx[((++spk)&31)] ^= enc_alpha_to[feedback +  30];
                  par_idx[((++spk)&31)] ^= enc_alpha_to[feedback +  97];
                  par_idx[((++spk)&31)] ^= enc_alpha_to[feedback + 251];
                  par_idx[((++spk)&31)] ^= enc_alpha_to[feedback + 126];
                  par_idx[((++spk)&31)] ^= enc_alpha_to[feedback +  43];
                  par_idx[((++spk)&31)] ^= enc_alpha_to[feedback +   4];
                  par_idx[((++spk)&31)] ^= enc_alpha_to[feedback +  66];
                  par_idx[((++spk)&31)] ^= enc_alpha_to[feedback +  59];
                  par_idx[((++spk)&31)] ^= enc_alpha_to[feedback + 249];

		  par_idx[sp] = enc_alpha_to[feedback];  /* feedback + 0 */
	       }
	       else                   /* zero feedback term */
		 par_idx[sp] = 0;

	       par_idx += nroots;
	    }

	    sp = (sp+1) & 31;         /* shift */

	    /* Report progress */

	    progress++;
	    percent = (1000*progress)/max_percent;
	    if(last_percent != percent) 
	    {  PrintMsg(_("Ecc generation: %3d.%1d%%\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b"),
		        percent/10, percent%10);
	       last_percent = percent;
	    }
	 }
	}
	break;

        case HIGH: /* Inner loop is unrolled for nroots = 64. */
	{int sp=1; /* sp==1 makes sure sp==0 after ndata bytes [since (191+1) mod 64 = 0] */
         for(layer=0; layer<ndata; layer++)
	 {  int offset = 0;
	    unsigned char *par_idx = parity;

	    /* Read the next data sectors of this layer. */

	    for(i=0; i<actual_layer_sectors; i++)
	    {  read_sector(ii, &eh, data+offset, block_idx[layer]);
	       block_idx[layer]++;
	       offset += 2048;
	    }

	    /* Now process the data bytes of the current layer. */

	    for(i=0; i<actual_layer_bytes; i++)
	    {  register int feedback;

	       feedback = gf_index_of[data[i] ^ par_idx[sp]];

	       if(feedback != ALPHA0) /* non-zero feedback term */
	       {  register int spk = sp;

	          par_idx[((++spk)&63)] ^= enc_alpha_to[feedback + gf_gpoly[63]];
		  par_idx[((++spk)&63)] ^= enc_alpha_to[feedback + gf_gpoly[62]];
		  par_idx[((++spk)&63)] ^= enc_alpha_to[feedback + gf_gpoly[61]];
		  par_idx[((++spk)&63)] ^= enc_alpha_to[feedback + gf_gpoly[60]];
		  par_idx[((++spk)&63)] ^= enc_alpha_to[feedback + gf_gpoly[59]];
		  par_idx[((++spk)&63)] ^= enc_alpha_to[feedback + gf_gpoly[58]];
		  par_idx[((++spk)&63)] ^= enc_alpha_to[feedback + gf_gpoly[57]];
		  par_idx[((++spk)&63)] ^= enc_alpha_to[feedback + gf_gpoly[56]];
		  par_idx[((++spk)&63)] ^= enc_alpha_to[feedback + gf_gpoly[55]];
		  par_idx[((++spk)&63)] ^= enc_alpha_to[feedback + gf_gpoly[54]];
		  par_idx[((++spk)&63)] ^= enc_alpha_to[feedback + gf_gpoly[53]];
		  par_idx[((++spk)&63)] ^= enc_alpha_to[feedback + gf_gpoly[52]];
		  par_idx[((++spk)&63)] ^= enc_alpha_to[feedback + gf_gpoly[51]];
		  par_idx[((++spk)&63)] ^= enc_alpha_to[feedback + gf_gpoly[50]];
		  par_idx[((++spk)&63)] ^= enc_alpha_to[feedback + gf_gpoly[49]];
		  par_idx[((++spk)&63)] ^= enc_alpha_to[feedback + gf_gpoly[48]];
		  par_idx[((++spk)&63)] ^= enc_alpha_to[feedback + gf_gpoly[47]];
		  par_idx[((++spk)&63)] ^= enc_alpha_to[feedback + gf_gpoly[46]];
		  par_idx[((++spk)&63)] ^= enc_alpha_to[feedback + gf_gpoly[45]];
		  par_idx[((++spk)&63)] ^= enc_alpha_to[feedback + gf_gpoly[44]];
		  par_idx[((++spk)&63)] ^= enc_alpha_to[feedback + gf_gpoly[43]];
		  par_idx[((++spk)&63)] ^= enc_alpha_to[feedback + gf_gpoly[42]];
		  par_idx[((++spk)&63)] ^= enc_alpha_to[feedback + gf_gpoly[41]];
		  par_idx[((++spk)&63)] ^= enc_alpha_to[feedback + gf_gpoly[40]];
		  par_idx[((++spk)&63)] ^= enc_alpha_to[feedback + gf_gpoly[39]];
		  par_idx[((++spk)&63)] ^= enc_alpha_to[feedback + gf_gpoly[38]];
		  par_idx[((++spk)&63)] ^= enc_alpha_to[feedback + gf_gpoly[37]];
		  par_idx[((++spk)&63)] ^= enc_alpha_to[feedback + gf_gpoly[36]];
		  par_idx[((++spk)&63)] ^= enc_alpha_to[feedback + gf_gpoly[35]];
		  par_idx[((++spk)&63)] ^= enc_alpha_to[feedback + gf_gpoly[34]];
		  par_idx[((++spk)&63)] ^= enc_alpha_to[feedback + gf_gpoly[33]];
		  par_idx[((++spk)&63)] ^= enc_alpha_to[feedback + gf_gpoly[32]];
		  par_idx[((++spk)&63)] ^= enc_alpha_to[feedback + gf_gpoly[31]];
		  par_idx[((++spk)&63)] ^= enc_alpha_to[feedback + gf_gpoly[30]];
		  par_idx[((++spk)&63)] ^= enc_alpha_to[feedback + gf_gpoly[29]];
		  par_idx[((++spk)&63)] ^= enc_alpha_to[feedback + gf_gpoly[28]];
		  par_idx[((++spk)&63)] ^= enc_alpha_to[feedback + gf_gpoly[27]];
		  par_idx[((++spk)&63)] ^= enc_alpha_to[feedback + gf_gpoly[26]];
		  par_idx[((++spk)&63)] ^= enc_alpha_to[feedback + gf_gpoly[25]];
		  par_idx[((++spk)&63)] ^= enc_alpha_to[feedback + gf_gpoly[24]];
		  par_idx[((++spk)&63)] ^= enc_alpha_to[feedback + gf_gpoly[23]];
		  par_idx[((++spk)&63)] ^= enc_alpha_to[feedback + gf_gpoly[22]];
		  par_idx[((++spk)&63)] ^= enc_alpha_to[feedback + gf_gpoly[21]];
		  par_idx[((++spk)&63)] ^= enc_alpha_to[feedback + gf_gpoly[20]];
		  par_idx[((++spk)&63)] ^= enc_alpha_to[feedback + gf_gpoly[19]];
		  par_idx[((++spk)&63)] ^= enc_alpha_to[feedback + gf_gpoly[18]];
		  par_idx[((++spk)&63)] ^= enc_alpha_to[feedback + gf_gpoly[17]];
		  par_idx[((++spk)&63)] ^= enc_alpha_to[feedback + gf_gpoly[16]];
		  par_idx[((++spk)&63)] ^= enc_alpha_to[feedback + gf_gpoly[15]];
		  par_idx[((++spk)&63)] ^= enc_alpha_to[feedback + gf_gpoly[14]];
		  par_idx[((++spk)&63)] ^= enc_alpha_to[feedback + gf_gpoly[13]];
		  par_idx[((++spk)&63)] ^= enc_alpha_to[feedback + gf_gpoly[12]];
		  par_idx[((++spk)&63)] ^= enc_alpha_to[feedback + gf_gpoly[11]];
		  par_idx[((++spk)&63)] ^= enc_alpha_to[feedback + gf_gpoly[10]];
		  par_idx[((++spk)&63)] ^= enc_alpha_to[feedback + gf_gpoly[ 9]];
		  par_idx[((++spk)&63)] ^= enc_alpha_to[feedback + gf_gpoly[ 8]];
		  par_idx[((++spk)&63)] ^= enc_alpha_to[feedback + gf_gpoly[ 7]];
		  par_idx[((++spk)&63)] ^= enc_alpha_to[feedback + gf_gpoly[ 6]];
		  par_idx[((++spk)&63)] ^= enc_alpha_to[feedback + gf_gpoly[ 5]];
		  par_idx[((++spk)&63)] ^= enc_alpha_to[feedback + gf_gpoly[ 4]];
		  par_idx[((++spk)&63)] ^= enc_alpha_to[feedback + gf_gpoly[ 3]];
		  par_idx[((++spk)&63)] ^= enc_alpha_to[feedback + gf_gpoly[ 2]];
		  par_idx[((++spk)&63)] ^= enc_alpha_to[feedback + gf_gpoly[ 1]];

		  par_idx[sp] = enc_alpha_to[feedback + gf_gpoly[0]];
	       }
	       else                   /* zero feedback term */
		 par_idx[sp] = 0;

	       par_idx += nroots;
	    }

	    sp = (sp+1) & 63;         /* shift */

	    /* Report progress */

	    progress++;
	    percent = (1000*progress)/max_percent;
	    if(last_percent != percent) 
	    {  PrintMsg(_("Ecc generation: %3d.%1d%%\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b"),
		        percent/10, percent%10);
	       last_percent = percent;
	    }
	 }
	}
	break;

        default:   /* general case for nroots other than 32 or 64 */
	{int sp = nroots - ndata % nroots; /* => (ndata + sp) mod nroots = 0 so that parity */
	                                   /* is aligned at sp=0 after ndata iterations */
      	 if(sp==nroots) sp=0;

	 for(layer=0; layer<ndata; layer++)
	 {  int offset = 0;
            unsigned char *par_idx = parity;

            /* Read the next data sectors of this layer. */

   	    for(i=0; i<actual_layer_sectors; i++)
	    {  read_sector(ii, &eh, data+offset, block_idx[layer]);
	       block_idx[layer]++;
	       offset += 2048;
	    }

	    /* Now process the data bytes of the current layer. */

	    for(i=0; i<actual_layer_bytes; i++)
	    {  register int feedback;

	       feedback = gf_index_of[data[i] ^ par_idx[sp]];

	       if(feedback != ALPHA0) /* non-zero feedback term */
	       {  register int spk = sp+1;
		  register int *gpoly = gf_gpoly + nroots;

		  switch(nroots-spk)
		  {  
		    case 110: par_idx[spk++] ^= enc_alpha_to[feedback + *--gpoly];
		    case 109: par_idx[spk++] ^= enc_alpha_to[feedback + *--gpoly];
		    case 108: par_idx[spk++] ^= enc_alpha_to[feedback + *--gpoly];
		    case 107: par_idx[spk++] ^= enc_alpha_to[feedback + *--gpoly];
		    case 106: par_idx[spk++] ^= enc_alpha_to[feedback + *--gpoly];
		    case 105: par_idx[spk++] ^= enc_alpha_to[feedback + *--gpoly];
		    case 104: par_idx[spk++] ^= enc_alpha_to[feedback + *--gpoly];
		    case 103: par_idx[spk++] ^= enc_alpha_to[feedback + *--gpoly];
		    case 102: par_idx[spk++] ^= enc_alpha_to[feedback + *--gpoly];
		    case 101: par_idx[spk++] ^= enc_alpha_to[feedback + *--gpoly];
		    case 100: par_idx[spk++] ^= enc_alpha_to[feedback + *--gpoly];
		     case 99: par_idx[spk++] ^= enc_alpha_to[feedback + *--gpoly];
		     case 98: par_idx[spk++] ^= enc_alpha_to[feedback + *--gpoly];
		     case 97: par_idx[spk++] ^= enc_alpha_to[feedback + *--gpoly];
		     case 96: par_idx[spk++] ^= enc_alpha_to[feedback + *--gpoly];
		     case 95: par_idx[spk++] ^= enc_alpha_to[feedback + *--gpoly];
		     case 94: par_idx[spk++] ^= enc_alpha_to[feedback + *--gpoly];
		     case 93: par_idx[spk++] ^= enc_alpha_to[feedback + *--gpoly];
		     case 92: par_idx[spk++] ^= enc_alpha_to[feedback + *--gpoly];
		     case 91: par_idx[spk++] ^= enc_alpha_to[feedback + *--gpoly];
		     case 90: par_idx[spk++] ^= enc_alpha_to[feedback + *--gpoly];
		     case 89: par_idx[spk++] ^= enc_alpha_to[feedback + *--gpoly];
		     case 88: par_idx[spk++] ^= enc_alpha_to[feedback + *--gpoly];
		     case 87: par_idx[spk++] ^= enc_alpha_to[feedback + *--gpoly];
		     case 86: par_idx[spk++] ^= enc_alpha_to[feedback + *--gpoly];
		     case 85: par_idx[spk++] ^= enc_alpha_to[feedback + *--gpoly];
		     case 84: par_idx[spk++] ^= enc_alpha_to[feedback + *--gpoly];
		     case 83: par_idx[spk++] ^= enc_alpha_to[feedback + *--gpoly];
		     case 82: par_idx[spk++] ^= enc_alpha_to[feedback + *--gpoly];
		     case 81: par_idx[spk++] ^= enc_alpha_to[feedback + *--gpoly];
		     case 80: par_idx[spk++] ^= enc_alpha_to[feedback + *--gpoly];
		     case 79: par_idx[spk++] ^= enc_alpha_to[feedback + *--gpoly];
		     case 78: par_idx[spk++] ^= enc_alpha_to[feedback + *--gpoly];
		     case 77: par_idx[spk++] ^= enc_alpha_to[feedback + *--gpoly];
		     case 76: par_idx[spk++] ^= enc_alpha_to[feedback + *--gpoly];
		     case 75: par_idx[spk++] ^= enc_alpha_to[feedback + *--gpoly];
		     case 74: par_idx[spk++] ^= enc_alpha_to[feedback + *--gpoly];
		     case 73: par_idx[spk++] ^= enc_alpha_to[feedback + *--gpoly];
		     case 72: par_idx[spk++] ^= enc_alpha_to[feedback + *--gpoly];
		     case 71: par_idx[spk++] ^= enc_alpha_to[feedback + *--gpoly];
		     case 70: par_idx[spk++] ^= enc_alpha_to[feedback + *--gpoly];
		     case 69: par_idx[spk++] ^= enc_alpha_to[feedback + *--gpoly];
		     case 68: par_idx[spk++] ^= enc_alpha_to[feedback + *--gpoly];
		     case 67: par_idx[spk++] ^= enc_alpha_to[feedback + *--gpoly];
		     case 66: par_idx[spk++] ^= enc_alpha_to[feedback + *--gpoly];
		     case 65: par_idx[spk++] ^= enc_alpha_to[feedback + *--gpoly];
		     case 64: par_idx[spk++] ^= enc_alpha_to[feedback + *--gpoly];
		     case 63: par_idx[spk++] ^= enc_alpha_to[feedback + *--gpoly];
		     case 62: par_idx[spk++] ^= enc_alpha_to[feedback + *--gpoly];
		     case 61: par_idx[spk++] ^= enc_alpha_to[feedback + *--gpoly];
		     case 60: par_idx[spk++] ^= enc_alpha_to[feedback + *--gpoly];
		     case 59: par_idx[spk++] ^= enc_alpha_to[feedback + *--gpoly];
		     case 58: par_idx[spk++] ^= enc_alpha_to[feedback + *--gpoly];
		     case 57: par_idx[spk++] ^= enc_alpha_to[feedback + *--gpoly];
		     case 56: par_idx[spk++] ^= enc_alpha_to[feedback + *--gpoly];
		     case 55: par_idx[spk++] ^= enc_alpha_to[feedback + *--gpoly];
		     case 54: par_idx[spk++] ^= enc_alpha_to[feedback + *--gpoly];
		     case 53: par_idx[spk++] ^= enc_alpha_to[feedback + *--gpoly];
		     case 52: par_idx[spk++] ^= enc_alpha_to[feedback + *--gpoly];
		     case 51: par_idx[spk++] ^= enc_alpha_to[feedback + *--gpoly];
		     case 50: par_idx[spk++] ^= enc_alpha_to[feedback + *--gpoly];
		     case 49: par_idx[spk++] ^= enc_alpha_to[feedback + *--gpoly];
		     case 48: par_idx[spk++] ^= enc_alpha_to[feedback + *--gpoly];
		     case 47: par_idx[spk++] ^= enc_alpha_to[feedback + *--gpoly];
		     case 46: par_idx[spk++] ^= enc_alpha_to[feedback + *--gpoly];
		     case 45: par_idx[spk++] ^= enc_alpha_to[feedback + *--gpoly];
		     case 44: par_idx[spk++] ^= enc_alpha_to[feedback + *--gpoly];
		     case 43: par_idx[spk++] ^= enc_alpha_to[feedback + *--gpoly];
		     case 42: par_idx[spk++] ^= enc_alpha_to[feedback + *--gpoly];
		     case 41: par_idx[spk++] ^= enc_alpha_to[feedback + *--gpoly];
		     case 40: par_idx[spk++] ^= enc_alpha_to[feedback + *--gpoly];
		     case 39: par_idx[spk++] ^= enc_alpha_to[feedback + *--gpoly];
		     case 38: par_idx[spk++] ^= enc_alpha_to[feedback + *--gpoly];
		     case 37: par_idx[spk++] ^= enc_alpha_to[feedback + *--gpoly];
		     case 36: par_idx[spk++] ^= enc_alpha_to[feedback + *--gpoly];
		     case 35: par_idx[spk++] ^= enc_alpha_to[feedback + *--gpoly];
		     case 34: par_idx[spk++] ^= enc_alpha_to[feedback + *--gpoly];
		     case 33: par_idx[spk++] ^= enc_alpha_to[feedback + *--gpoly];
		     case 32: par_idx[spk++] ^= enc_alpha_to[feedback + *--gpoly];
		     case 31: par_idx[spk++] ^= enc_alpha_to[feedback + *--gpoly];
		     case 30: par_idx[spk++] ^= enc_alpha_to[feedback + *--gpoly];
		     case 29: par_idx[spk++] ^= enc_alpha_to[feedback + *--gpoly];
		     case 28: par_idx[spk++] ^= enc_alpha_to[feedback + *--gpoly];
		     case 27: par_idx[spk++] ^= enc_alpha_to[feedback + *--gpoly];
		     case 26: par_idx[spk++] ^= enc_alpha_to[feedback + *--gpoly];
		     case 25: par_idx[spk++] ^= enc_alpha_to[feedback + *--gpoly];
		     case 24: par_idx[spk++] ^= enc_alpha_to[feedback + *--gpoly];
		     case 23: par_idx[spk++] ^= enc_alpha_to[feedback + *--gpoly];
		     case 22: par_idx[spk++] ^= enc_alpha_to[feedback + *--gpoly];
		     case 21: par_idx[spk++] ^= enc_alpha_to[feedback + *--gpoly];
		     case 20: par_idx[spk++] ^= enc_alpha_to[feedback + *--gpoly];
		     case 19: par_idx[spk++] ^= enc_alpha_to[feedback + *--gpoly];
		     case 18: par_idx[spk++] ^= enc_alpha_to[feedback + *--gpoly];
		     case 17: par_idx[spk++] ^= enc_alpha_to[feedback + *--gpoly];
		     case 16: par_idx[spk++] ^= enc_alpha_to[feedback + *--gpoly];
		     case 15: par_idx[spk++] ^= enc_alpha_to[feedback + *--gpoly];
		     case 14: par_idx[spk++] ^= enc_alpha_to[feedback + *--gpoly];
		     case 13: par_idx[spk++] ^= enc_alpha_to[feedback + *--gpoly];
		     case 12: par_idx[spk++] ^= enc_alpha_to[feedback + *--gpoly];
		     case 11: par_idx[spk++] ^= enc_alpha_to[feedback + *--gpoly];
		     case 10: par_idx[spk++] ^= enc_alpha_to[feedback + *--gpoly];
		     case  9: par_idx[spk++] ^= enc_alpha_to[feedback + *--gpoly];
		     case  8: par_idx[spk++] ^= enc_alpha_to[feedback + *--gpoly];
		     case  7: par_idx[spk++] ^= enc_alpha_to[feedback + *--gpoly];
		     case  6: par_idx[spk++] ^= enc_alpha_to[feedback + *--gpoly];
		     case  5: par_idx[spk++] ^= enc_alpha_to[feedback + *--gpoly];
		     case  4: par_idx[spk++] ^= enc_alpha_to[feedback + *--gpoly];
		     case  3: par_idx[spk++] ^= enc_alpha_to[feedback + *--gpoly];
		     case  2: par_idx[spk++] ^= enc_alpha_to[feedback + *--gpoly];
		     case  1: par_idx[spk++] ^= enc_alpha_to[feedback + *--gpoly];
		  }

		  spk = 0;
		  
		  switch(sp)
		  {
                    case 110: par_idx[spk++] ^= enc_alpha_to[feedback + *--gpoly];
		    case 109: par_idx[spk++] ^= enc_alpha_to[feedback + *--gpoly];
		    case 108: par_idx[spk++] ^= enc_alpha_to[feedback + *--gpoly];
		    case 107: par_idx[spk++] ^= enc_alpha_to[feedback + *--gpoly];
		    case 106: par_idx[spk++] ^= enc_alpha_to[feedback + *--gpoly];
		    case 105: par_idx[spk++] ^= enc_alpha_to[feedback + *--gpoly];
		    case 104: par_idx[spk++] ^= enc_alpha_to[feedback + *--gpoly];
		    case 103: par_idx[spk++] ^= enc_alpha_to[feedback + *--gpoly];
		    case 102: par_idx[spk++] ^= enc_alpha_to[feedback + *--gpoly];
		    case 101: par_idx[spk++] ^= enc_alpha_to[feedback + *--gpoly];
		    case 100: par_idx[spk++] ^= enc_alpha_to[feedback + *--gpoly];
		     case 99: par_idx[spk++] ^= enc_alpha_to[feedback + *--gpoly];
		     case 98: par_idx[spk++] ^= enc_alpha_to[feedback + *--gpoly];
		     case 97: par_idx[spk++] ^= enc_alpha_to[feedback + *--gpoly];
		     case 96: par_idx[spk++] ^= enc_alpha_to[feedback + *--gpoly];
		     case 95: par_idx[spk++] ^= enc_alpha_to[feedback + *--gpoly];
		     case 94: par_idx[spk++] ^= enc_alpha_to[feedback + *--gpoly];
		     case 93: par_idx[spk++] ^= enc_alpha_to[feedback + *--gpoly];
		     case 92: par_idx[spk++] ^= enc_alpha_to[feedback + *--gpoly];
		     case 91: par_idx[spk++] ^= enc_alpha_to[feedback + *--gpoly];
		     case 90: par_idx[spk++] ^= enc_alpha_to[feedback + *--gpoly];
		     case 89: par_idx[spk++] ^= enc_alpha_to[feedback + *--gpoly];
		     case 88: par_idx[spk++] ^= enc_alpha_to[feedback + *--gpoly];
		     case 87: par_idx[spk++] ^= enc_alpha_to[feedback + *--gpoly];
		     case 86: par_idx[spk++] ^= enc_alpha_to[feedback + *--gpoly];
		     case 85: par_idx[spk++] ^= enc_alpha_to[feedback + *--gpoly];
		     case 84: par_idx[spk++] ^= enc_alpha_to[feedback + *--gpoly];
		     case 83: par_idx[spk++] ^= enc_alpha_to[feedback + *--gpoly];
		     case 82: par_idx[spk++] ^= enc_alpha_to[feedback + *--gpoly];
		     case 81: par_idx[spk++] ^= enc_alpha_to[feedback + *--gpoly];
		     case 80: par_idx[spk++] ^= enc_alpha_to[feedback + *--gpoly];
		     case 79: par_idx[spk++] ^= enc_alpha_to[feedback + *--gpoly];
		     case 78: par_idx[spk++] ^= enc_alpha_to[feedback + *--gpoly];
		     case 77: par_idx[spk++] ^= enc_alpha_to[feedback + *--gpoly];
		     case 76: par_idx[spk++] ^= enc_alpha_to[feedback + *--gpoly];
		     case 75: par_idx[spk++] ^= enc_alpha_to[feedback + *--gpoly];
		     case 74: par_idx[spk++] ^= enc_alpha_to[feedback + *--gpoly];
		     case 73: par_idx[spk++] ^= enc_alpha_to[feedback + *--gpoly];
		     case 72: par_idx[spk++] ^= enc_alpha_to[feedback + *--gpoly];
		     case 71: par_idx[spk++] ^= enc_alpha_to[feedback + *--gpoly];
		     case 70: par_idx[spk++] ^= enc_alpha_to[feedback + *--gpoly];
		     case 69: par_idx[spk++] ^= enc_alpha_to[feedback + *--gpoly];
		     case 68: par_idx[spk++] ^= enc_alpha_to[feedback + *--gpoly];
		     case 67: par_idx[spk++] ^= enc_alpha_to[feedback + *--gpoly];
		     case 66: par_idx[spk++] ^= enc_alpha_to[feedback + *--gpoly];
		     case 65: par_idx[spk++] ^= enc_alpha_to[feedback + *--gpoly];
		     case 64: par_idx[spk++] ^= enc_alpha_to[feedback + *--gpoly];
		     case 63: par_idx[spk++] ^= enc_alpha_to[feedback + *--gpoly];
		     case 62: par_idx[spk++] ^= enc_alpha_to[feedback + *--gpoly];
		     case 61: par_idx[spk++] ^= enc_alpha_to[feedback + *--gpoly];
		     case 60: par_idx[spk++] ^= enc_alpha_to[feedback + *--gpoly];
		     case 59: par_idx[spk++] ^= enc_alpha_to[feedback + *--gpoly];
		     case 58: par_idx[spk++] ^= enc_alpha_to[feedback + *--gpoly];
		     case 57: par_idx[spk++] ^= enc_alpha_to[feedback + *--gpoly];
		     case 56: par_idx[spk++] ^= enc_alpha_to[feedback + *--gpoly];
		     case 55: par_idx[spk++] ^= enc_alpha_to[feedback + *--gpoly];
		     case 54: par_idx[spk++] ^= enc_alpha_to[feedback + *--gpoly];
		     case 53: par_idx[spk++] ^= enc_alpha_to[feedback + *--gpoly];
		     case 52: par_idx[spk++] ^= enc_alpha_to[feedback + *--gpoly];
		     case 51: par_idx[spk++] ^= enc_alpha_to[feedback + *--gpoly];
		     case 50: par_idx[spk++] ^= enc_alpha_to[feedback + *--gpoly];
		     case 49: par_idx[spk++] ^= enc_alpha_to[feedback + *--gpoly];
		     case 48: par_idx[spk++] ^= enc_alpha_to[feedback + *--gpoly];
		     case 47: par_idx[spk++] ^= enc_alpha_to[feedback + *--gpoly];
		     case 46: par_idx[spk++] ^= enc_alpha_to[feedback + *--gpoly];
		     case 45: par_idx[spk++] ^= enc_alpha_to[feedback + *--gpoly];
		     case 44: par_idx[spk++] ^= enc_alpha_to[feedback + *--gpoly];
		     case 43: par_idx[spk++] ^= enc_alpha_to[feedback + *--gpoly];
		     case 42: par_idx[spk++] ^= enc_alpha_to[feedback + *--gpoly];
		     case 41: par_idx[spk++] ^= enc_alpha_to[feedback + *--gpoly];
		     case 40: par_idx[spk++] ^= enc_alpha_to[feedback + *--gpoly];
		     case 39: par_idx[spk++] ^= enc_alpha_to[feedback + *--gpoly];
		     case 38: par_idx[spk++] ^= enc_alpha_to[feedback + *--gpoly];
		     case 37: par_idx[spk++] ^= enc_alpha_to[feedback + *--gpoly];
		     case 36: par_idx[spk++] ^= enc_alpha_to[feedback + *--gpoly];
		     case 35: par_idx[spk++] ^= enc_alpha_to[feedback + *--gpoly];
		     case 34: par_idx[spk++] ^= enc_alpha_to[feedback + *--gpoly];
		     case 33: par_idx[spk++] ^= enc_alpha_to[feedback + *--gpoly];
		     case 32: par_idx[spk++] ^= enc_alpha_to[feedback + *--gpoly];
		     case 31: par_idx[spk++] ^= enc_alpha_to[feedback + *--gpoly];
		     case 30: par_idx[spk++] ^= enc_alpha_to[feedback + *--gpoly];
		     case 29: par_idx[spk++] ^= enc_alpha_to[feedback + *--gpoly];
		     case 28: par_idx[spk++] ^= enc_alpha_to[feedback + *--gpoly];
		     case 27: par_idx[spk++] ^= enc_alpha_to[feedback + *--gpoly];
		     case 26: par_idx[spk++] ^= enc_alpha_to[feedback + *--gpoly];
		     case 25: par_idx[spk++] ^= enc_alpha_to[feedback + *--gpoly];
		     case 24: par_idx[spk++] ^= enc_alpha_to[feedback + *--gpoly];
		     case 23: par_idx[spk++] ^= enc_alpha_to[feedback + *--gpoly];
		     case 22: par_idx[spk++] ^= enc_alpha_to[feedback + *--gpoly];
		     case 21: par_idx[spk++] ^= enc_alpha_to[feedback + *--gpoly];
		     case 20: par_idx[spk++] ^= enc_alpha_to[feedback + *--gpoly];
		     case 19: par_idx[spk++] ^= enc_alpha_to[feedback + *--gpoly];
		     case 18: par_idx[spk++] ^= enc_alpha_to[feedback + *--gpoly];
		     case 17: par_idx[spk++] ^= enc_alpha_to[feedback + *--gpoly];
		     case 16: par_idx[spk++] ^= enc_alpha_to[feedback + *--gpoly];
		     case 15: par_idx[spk++] ^= enc_alpha_to[feedback + *--gpoly];
		     case 14: par_idx[spk++] ^= enc_alpha_to[feedback + *--gpoly];
		     case 13: par_idx[spk++] ^= enc_alpha_to[feedback + *--gpoly];
		     case 12: par_idx[spk++] ^= enc_alpha_to[feedback + *--gpoly];
		     case 11: par_idx[spk++] ^= enc_alpha_to[feedback + *--gpoly];
		     case 10: par_idx[spk++] ^= enc_alpha_to[feedback + *--gpoly];
		     case  9: par_idx[spk++] ^= enc_alpha_to[feedback + *--gpoly];
		     case  8: par_idx[spk++] ^= enc_alpha_to[feedback + *--gpoly];
		     case  7: par_idx[spk++] ^= enc_alpha_to[feedback + *--gpoly];
		     case  6: par_idx[spk++] ^= enc_alpha_to[feedback + *--gpoly];
		     case  5: par_idx[spk++] ^= enc_alpha_to[feedback + *--gpoly];
		     case  4: par_idx[spk++] ^= enc_alpha_to[feedback + *--gpoly];
		     case  3: par_idx[spk++] ^= enc_alpha_to[feedback + *--gpoly];
		     case  2: par_idx[spk++] ^= enc_alpha_to[feedback + *--gpoly];
		     case  1: par_idx[spk++] ^= enc_alpha_to[feedback + *--gpoly];
		  }

		  par_idx[sp] = enc_alpha_to[feedback + gf_gpoly[0]];
	       }
	       else                   /* zero feedback term */
		 par_idx[sp] = 0;

	       par_idx += nroots;
	    }

	    if(++sp>=nroots) sp=0;   /* shift */

	    /* Report progress */

	    progress++;
	    percent = (1000*progress)/max_percent;
	    if(last_percent != percent) 
	    {  PrintMsg(_("Ecc generation: %3d.%1d%%\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b"),
		        percent/10, percent%10);
	       last_percent = percent;
	    }
	 }
	}
	break;
      }

      /* Write the nroots bytes of parity information */

      n = fwrite(parity, nroots, actual_layer_bytes, ecc);

      if(n != actual_layer_bytes)
        Stop(_("could not write to ecc file \"%s\":\n%s"),eccname,strerror(errno));

      MD5Update(&ii->md5Ctxt, parity, nroots*actual_layer_bytes);
   }

   /*** Complete the ecc header and write it out */

   MD5Final(eh.eccSum, &ii->md5Ctxt);

   rewind(ecc);
   n = fwrite(&eh, sizeof(EccHeader), 1, ecc);
   if(n != 1)
     Stop(_("Can't write ecc header:\n%s"),strerror(errno));

   PrintMsg(_("Ecc generation: 100.0%%\n"
	      "Error correction file \"%s\" created.\n"
	      "Make sure to keep this file on a reliable medium.\n"),
	      eccname);

   /*** Clean up */
      
   g_free(data);
   g_free(parity);

   free_tables();

#ifdef DEBUG_READ
   for(i=0; i<(int)ii->sectors; i++)
     if(bucket[i] != 1)
       PrintMsg("sector %d: %d reads\n",i,bucket[i]);

   PrintMsg("Sectors: %d expected, %d dead, %d padded, %d read.\n",(int)ii->sectors ,deadcnt,padcnt,seccnt); 
#endif
}
