/*  dvdisaster: Additional error correction for optical media.
 *  Copyright (C) 2004,2005 Carsten Gnoerlich.
 *  Project home page: http://www.dvdisaster.com
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA,
 *  or direct your browser at http://www.gnu.org.
 */

#include "dvdisaster.h"

/*
 * Adapt the Windows approach to large files
 * to the POSIX way.
 */

#ifdef SYS_MINGW

#include <windows.h>

int fseeko(FILE *file, gint64 pos, int mode)
{  
   if(mode != SEEK_SET) Stop(_("fseeko mode must be SEEK_SET"));

   return fsetpos(file, &pos);
}

int ftruncate(int fd, gint64 size)
{  gint32 handle;

   if((handle=_get_osfhandle(fd)) == -1)
     return -1;

   if(_lseeki64(fd, size, SEEK_SET) == -1)
     return -1;

   if(SetEndOfFile((HANDLE)handle) == 0)
     return -1;

   return 0;
}

#endif

/*
 * The usual gettext() wrapper so that we can use strings
 * like "file!quit" in menus.
 */

#ifdef WITH_NLS_YES
char *sgettext(char *msgid)
{  char *msgval = gettext(msgid);
   
   if(msgval == msgid)
   {  msgval = strrchr(msgid, '|');
      if(msgval) return msgval + 1;
      else       return msgid;
   }
   return msgval;
}
#endif

/*
 * Convert between 8 bytes of unsigned char and gint64.
 * Mainly needed as various version of gcc used different alignment for gint64 values
 * so that we can not use gint64 directly in the EccHeader structure.
 */

gint64 uchar_to_gint64(unsigned char *bytes)
{  gint64 out;

   memcpy(&out, bytes, sizeof(gint64));

   return out;
}

void gint64_to_uchar(unsigned char *out, gint64 in)
{
   memcpy(out, &in, sizeof(in));
}

/*
 * Print high- and low-priority messages to the console,
 * so that user can sort them out using shell redirection.
 *
 * PrintMsg() writes to stderr, while
 * PrintLog() writes to stdout.
 *
 * This seems natural as >/dev/null will just filter out
 * the unimportant stuff.
 */

void PrintMsg(char *format, ...)
{  va_list argp;
  
   va_start(argp, format);
   g_vfprintf(stderr, format, argp);
   va_end(argp);

   fflush(stderr);   /* at least needed for Windows */
}

void PrintLog(char *format, ...)
{  va_list argp;
  
   va_start(argp, format);
   g_vprintf(format, argp);
   va_end(argp);

   fflush(stdout);   /* at least needed for Windows */
}

/*
 * Tell user that current action was aborted
 * due to a serious error.
 */

void Stop(char *format, ...)
{  va_list argp;

   g_fprintf(stderr, _("*\n* dvdisaster - can not continue:\n*\n"));
   va_start(argp, format);
   g_vfprintf(stderr, format, argp);
   va_end(argp);
   g_fprintf(stderr, "\n\n");
   fflush(stderr);

   exit(EXIT_FAILURE);
}

/*
 * Convert md5 digest into printable format.
 */

static char hextab[] = "0123456789abcdef";

void AsciiDigest(char *out, unsigned char *in)
{  int i,o;

   for(i=0, o=0; i<16; i++)
   {  out[o++] = hextab[in[i]>>4];
      out[o++] = hextab[in[i]&0x0f];
   }
   out[o] = 0;
}

/*
 * Calculate total number of sectors (including the last sector)
 * and the remaining bytes in the last sector for a given medium size.
 */

void CalcSectors(gint64 size, gint64 *sectors, int *in_last)
{
   *sectors = size/2048;
   *in_last = size & 0x7ff; 
   if(*in_last>0) (*sectors)++;
   else *in_last = 2048;
}
