/*  dvdisaster: Additional error correction for optical media.
 *  Copyright (C) 2004,2005 Carsten Gnoerlich.
 *  Project home page: http://www.dvdisaster.com
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA,
 *  or direct your browser at http://www.gnu.org.
 */

#include "dvdisaster.h"

#ifdef SYS_LINUX
#include <sys/ioctl.h>
#include <linux/cdrom.h>
#endif

#ifdef SYS_MINGW
#include <windows.h>
#include <winioctl.h>
#endif

/***
 *** OS specific wrappers for querying and reading the optical drive
 ***/

/*
 * Define the Sense data structure.
 */

#ifdef SYS_LINUX
typedef struct request_sense Sense;
#endif

#ifdef SYS_MINGW

/* This is actually the little endian version of the
 * Linux Sense struct. Windows uses only the first 24 bytes.
 */

typedef struct request_sense {
	guint8 error_code		: 7;
	guint8 valid			: 1;
	guint8 segment_number;
	guint8 sense_key		: 4;
	guint8 reserved2		: 1;
	guint8 ili			: 1;
	guint8 reserved1		: 2;
	guint8 information[4];
	guint8 add_sense_len;
	guint8 command_info[4];
	guint8 asc;
	guint8 ascq;
	guint8 fruc;
	guint8 sks[3];
	guint8 asb[46];
} Sense;
#endif

/*
 * Some collected info about the opened device and the inserted medium
 */

typedef struct _device_handle
{
#ifdef SYS_LINUX
   int fd;                    /* device file descriptor */
#endif
#ifdef SYS_MINGW
   HANDLE fd;                 /* Windows file handle for the device (SPTI case) */
   HINSTANCE aspiLib;         /* for opening the ASPI library */
   DWORD (*GetASPI32SupportInfo)(void);
   DWORD (*SendASPI32Command)(void*);
   int ha,target,lun;         /* ASPI way of describing drives */ 
#if 0
   BOOL  (*TranslateASPI32Address)(DWORD*, DWORD*);
#endif
#endif
   char *device;
   char devinfo[34];
   int (*read)(struct _device_handle*, char*, int, int);
   Sense sense;               /* sense data from last operation */
   gint64 userAreaSize;        /* size of user area according to DVD Info struct */
   int sessions;
   int layers;
   int mainType;
   int subType;
   char *typedescr;
} device_handle;

enum { DATA1, XA21, CD, DVD, UNSUPPORTED };

#define DATA_WRITE 0
#define DATA_READ  1
#define DATA_NONE  2

/**
 ** The Linux wrapper. Uses the uniform CDROM driver. Kernel 2.6.x recommended. 
 **/

#ifdef SYS_LINUX

#define MAX_CDB_SIZE CDROM_PACKET_SIZE

char* DefaultDevice()
{  
  return g_strdup("/dev/cdrom");
}

static device_handle* open_device(char *device)
{  device_handle *dh; 

   dh = g_malloc0(sizeof(device_handle));
   dh->fd = open(device, O_RDONLY | O_NONBLOCK);

   if(dh->fd < 0)
     Stop(_("Could not open %s: %s"),device, strerror(errno));

   dh->device = g_strdup(device);

   return dh;
}

static void close_device(device_handle *dh)
{
  close(dh->fd);
  g_free(dh->device);
  g_free(dh);
}

static int send_packet(device_handle *dh, char *cmd, int cdb_size, char *buf, int size, Sense *sense, int data_mode)
{  struct cdrom_generic_command cgc;

   memset(&cgc, 0, sizeof(cgc));

   memcpy(cgc.cmd, cmd, MAX_CDB_SIZE);  /* Linux ignores the CDB size */
   cgc.buffer = buf;
   cgc.buflen = size;
   cgc.sense  = sense;
   cgc.timeout = 5000;   /* 5 * HZ (from kernel's point of view) */

   switch(data_mode)
   {  case DATA_READ:
        cgc.data_direction = CGC_DATA_READ; 
	break;
      default:
	Stop(_("illegal data_mode: %d"),data_mode);
   }

   return ioctl(dh->fd, CDROM_SEND_PACKET, &cgc);
}
#endif /* SYS_LINUX */

/**
 ** Windows wrapper. 
 **
 * Actually we have two wrappers; one for SPTI and one for ASPI.
 * SPTI requires Windows 2000 or XP and root priviledges, but seems
 * to be more compatible.
 * Otoh, ASPI run without special priviledges and even on the older
 * Windows 9x versions. 
 * The APSI wrapper has only been tested against Adaptec ASPI so far.
 *
 * SPTI is tried first and then we fall back to ASPI.
 */

#ifdef SYS_MINGW

#define MAX_CDB_SIZE 16   /* longest possible SCSI command */

/*
 * This is independent from later decision between SPTI and ASPI. 
 * The drive letter ordering to remembered to aid the 
 * drive letter / ASPI drive guessing / mapping.
 */

static int drive_letters[26];
static int cd_dvd_drives = 0;

char* DefaultDevice()
{  static char picked[3] = "C:";
   char drive_letter;
   char device_path[16];

   for(drive_letter='C'; drive_letter<='Z'; drive_letter++)
   {  UINT drive_type;

      g_sprintf(device_path, "%c:\\", drive_letter);
      drive_type = GetDriveType(device_path);

      if(drive_type == DRIVE_CDROM)
      {  drive_letters[cd_dvd_drives] = drive_letter;
	 cd_dvd_drives++;

         if(cd_dvd_drives == 1)
	   *picked = drive_letter;
      }
   }

   if(cd_dvd_drives > 0)
     return picked;

   return "No_Drive_found";
}

/*
 * Close the SPTI/ASPI devices.
 */

static void close_device(device_handle *dh)
{
  if(dh->aspiLib)  /* Did we run in ASPI mode? */
  {  FreeLibrary(dh->aspiLib);
  }
  else             /* SPTI cleanup */
  {  CloseHandle(dh->fd);
  }

  g_free(dh->device);
  g_free(dh);
}

/**
 ** The ASPI wrapper.
 **/

/* 
 * Not everyone has the development kit with the #includes
 * for sending SCSI packets via ASPI, so we define what we need here.
 */

#define SS_PENDING        0x00
#define SS_COMP           0x01
#define SS_ERR            0x04

#define SRB_DIR_IN        0x08
#define SRB_DIR_OUT       0x10
#define SRB_EVENT_NOTIFY  0x40

#define SC_HA_INQUIRY     0x00 /* Host adapter inquiry */
#define SC_GET_DEV_TYPE   0x01 /* Get device type */
#define SC_EXEC_SCSI_CMD  0x02 /* Execute SCSI command */

#define DTYPE_DASD  0
#define DTYPE_CDROM 5

typedef struct
{   BYTE   Cmd;                /* ASPI command code = SC_HA_INQUIRY */
    BYTE   Status;             /* ASPI command status byte */
    BYTE   HaId;               /* ASPI host adapter number */
    BYTE   Flags;              /* ASPI request flags */
    DWORD  Hdr_Rsvd;           /* Reserved, MUST = 0 */
    BYTE   Count;              /* Number of host adapters present */
    BYTE   SCSI_ID;            /* SCSI ID of host adapter */
    BYTE   HA_ManagerId[16];   /* String describing the manager */
    BYTE   HA_Identifier[16];  /* String describing the host adapter */
    WORD   HA_BufAlignMask;    /* HA_Unique[1-0] */
    BYTE   HA_Flags;           /* HA_Unique[2] */
    BYTE   HA_MaxTargets;      /* HA_Unique[3] */
    DWORD  HA_MaxTransferLength; /* HA_Unique[7-4] */
    DWORD  HA_MaxSGElements;   /* HA_Unique[11-8] */
    BYTE   HA_Rsvd2[4];        /* HA_Unique[15-12] */
    WORD   HA_Rsvd1;           /* Reserved, MUST = 0 */
} PACKED SRB32_HAInquiry;

typedef struct
{   BYTE   Cmd;                /* ASPI command code = SC_GET_DEV_TYPE */
    BYTE   Status;             /* ASPI command status byte */
    BYTE   HaId;               /* ASPI host adapter number */
    BYTE   Flags;              /* Reserved, MUST = 0 */
    DWORD  Hdr_Rsvd;           /* Reserved, MUST = 0 */
    BYTE   Target;             /* Target's SCSI ID */
    BYTE   Lun;                /* Target's LUN number */
    BYTE   DeviceType;         /* Target's peripheral device type */
    BYTE   Rsvd1;              /* Reserved, MUST = 0 */
} PACKED SRB32_GDEVBlock;

typedef struct
{   BYTE   Cmd;                /* ASPI command code = SC_EXEC_SCSI_CMD */
    BYTE   Status;             /* ASPI command status byte */
    BYTE   HaId;               /* ASPI host adapter number */
    BYTE   Flags;              /* ASPI request flags */
    DWORD  Hdr_Rsvd;           /* Reserved */
    BYTE   Target;             /* Target's SCSI ID */
    BYTE   Lun;                /* Target's LUN number */
    WORD   Rsvd1;              /* Reserved for Alignment */
    DWORD  BufLen;             /* Data Allocation Length */
    BYTE   *BufPtr;            /* Data Buffer Pointer */
    BYTE   SenseLen;           /* Sense Allocation Length */
    BYTE   CDBLen;             /* CDB Length */
    BYTE   HaStat;             /* Host Adapter Status */
    BYTE   TargStat;           /* Target Status */
    VOID   *PostProc;          /* Post routine */
    BYTE   Rsvd2[20];          /* Reserved, MUST be 0 */
    BYTE   CDBByte[16];        /* SCSI CDB */
    BYTE   SenseArea[16];      /* Request Sense buffer */
} PACKED SRB32_ExecSCSICmd;

/*
 * Open and prepare the device using ASPI.
 */

static device_handle* open_aspi_device(char *device, int verbose)
{ device_handle *dh;
  int status,ret,ha,max_ha;
  SRB32_HAInquiry ha_inq;
  int drive_count = 0;
  int drive_wanted = 0;
  char spti_name[3];
  char letter_wanted = toupper(*device);

  dh = g_malloc0(sizeof(device_handle));
  dh->device = g_strdup(device);

  /* Look for our special ASPI drive syntax */

  if(*device >= '1' && *device <= '9')
    drive_wanted = *device - '0';

  /* Try to open the ASPI library */

  dh->aspiLib = LoadLibrary("WNASPI32.DLL");
  if(!dh->aspiLib)
  {  g_free(dh); 
     return NULL;
  }

  dh->GetASPI32SupportInfo 
     = (DWORD(*)(void))GetProcAddress(dh->aspiLib, "GetASPI32SupportInfo");
  if(!dh->GetASPI32SupportInfo)
  {  PrintMsg(_("GetASPI32SupportInfo() not available."));
     g_free(dh->device);
     g_free(dh); 
     return NULL;
  }

  dh->SendASPI32Command 
     = (DWORD(*)(void*))GetProcAddress(dh->aspiLib, "SendASPI32Command");
  if(!dh->SendASPI32Command)
  {  PrintMsg(_("SendASPI32Command() not available."));
     g_free(dh->device);
     g_free(dh); 
     return NULL;
  }

  /* The Adaptec docs seem to imply that this call is needed
     to initialize the ASPI library. Never mind, it is even slightly
     useful since it gives us the number of host adapters. */

  ret    = dh->GetASPI32SupportInfo();  
  status = (ret>>8) & 0xff;
  max_ha = ret & 0xff;

  if(status != SS_COMP)
  {  PrintMsg(_("Could not determine number of host adapters\n"));
     g_free(dh->device);
     g_free(dh); 
     return NULL;
  }

#if 0
  PrintMsg("Status %d, %d host adapters\n",status,max_ha);
#endif

  /* Now see if we can find any CDROM drives. */

  for(ha=0; ha<max_ha; ha++)
  {  int target,n_targets;

     memset(&ha_inq, 0, sizeof(SRB32_HAInquiry));
     ha_inq.Cmd  = SC_HA_INQUIRY;
     ha_inq.HaId = ha;

     dh->SendASPI32Command(&ha_inq);
     if(ha_inq.Status != SS_COMP)
       PrintMsg(_("ASPI warning: Could not query host adapter %d\n"),ha);

#if 0
     PrintMsg("HA %d: %16s\n",ha,ha_inq.HA_Identifier);
#endif

     if(ha_inq.HA_MaxTargets == 16)  /* my interpretation of */
          n_targets = 16;            /* Adaptecs documentation */
     else n_targets = 8;

#if 0
     /* Some ASPI drivers return -1 or 0 for actually working
	configurations, so we can`t rely on this information. */

     if(ha_inq.HA_MaxTransferLength == 0)
       continue;
#endif

     /* Missing the following conditions is close to impossible. */

     if(ha_inq.HA_BufAlignMask >= 4096)
       Stop(_("ASPI alignment = %d requested; can't handle that.\n"),
	     ha_inq.HA_BufAlignMask);
#if 0
     /* Some ASPI drivers return 0 or -1 for actually working drives,
	so this information is also useless. */

     if(ha_inq.HA_MaxTransferLength < 32768)
       Stop(_("ASPI max xfer length = %d; can't handle that.\n"),
	     ha_inq.HA_MaxTransferLength);
#endif
     
     /* Iterate over the HA's possible targets */

     for(target=0; target<n_targets; target++)
     {  SRB32_GDEVBlock gdb;

        if(target == ha_inq.SCSI_ID) /* ignore the HA itself */
          continue;

	memset(&gdb, 0, sizeof(SRB32_GDEVBlock));
	gdb.Cmd    = SC_GET_DEV_TYPE;
	gdb.HaId   = ha;
	gdb.Target = target;
	gdb.Lun    = 0;  /* Are there CDROMs with Lun != 0 ? */
 
	dh->SendASPI32Command(&gdb);

	if(gdb.Status != SS_COMP)
	  continue;  /* device does not exist */

	if(gdb.DeviceType == DTYPE_CDROM)
	{  char guessed_letter = drive_letters[drive_count];

	   drive_count++;

	   dh->ha = ha;
	   dh->target = target;
	   dh->lun = 0;

	   if(verbose)
	   { static void inquire_device(device_handle*); 

	     inquire_device(dh);
	     if(drive_count<26 && guessed_letter)
	     {  spti_name[0] = guessed_letter;
	        spti_name[1] = ':';
	     }
	     else spti_name[0] = spti_name[1] = '?';
	     spti_name[2] = 0;

	     PrintMsg(" %d: (%s) %s\n", drive_count, spti_name, dh->devinfo);
	   }

	   if(drive_wanted && drive_wanted == drive_count)
	     return dh;  /* drive found by our ASPI syntax */

	   if(letter_wanted == guessed_letter)
	     return dh;  /* hopefully found the right drive for letter */
	}
    }
  }

  if(verbose)
  {  if(!drive_count)
       PrintMsg(_("ASPI manager present, but no CD/DVD drives managed.\n"));
     else if(drive_count != cd_dvd_drives)
       PrintMsg(_("* Warning: %d SPTI drives, but %d ASPI drives.\n"
	          "           Drive letter mapping for ASPI drives is probably incorrect.\n"),
	        cd_dvd_drives, drive_count);
     return dh;
  }

  g_free(dh->device);
  g_free(dh); 
  return NULL;
}

/*
 * Print all CDROM drives accessible over ASPI.
 */

void ListAspiDrives()
{  device_handle *dh;

   PrintMsg(_("\nList of ASPI CD/DVD drives:\n")); 

   dh = open_aspi_device("A:", 1);

   if(dh) 
   {  PrintMsg(_("\nTo force ASPI usage over SPTI, refer to the drive by the\n"
	         "above numbers (use 1:, 2:,... instead of C:, D:,...)\n"));

      close_device(dh);   
   }
   else PrintMsg(_("ASPI manager not available or installed.\n"));
}


/*
 * Send the SCSI command through ASPI.
 */

static int send_aspi_packet(device_handle *dh, char *cmd, int cdb_size, char *buf, int size, Sense *sense, int data_mode)
{  SRB32_ExecSCSICmd srb;
   DWORD status;
   HANDLE srb_event;

   srb_event = CreateEvent(NULL, 1, 0, NULL);

   /* Prepare the SRB struct */

   memset(&srb, 0, sizeof(SRB32_ExecSCSICmd));
   srb.Cmd     = SC_EXEC_SCSI_CMD;
   srb.HaId    = dh->ha;
   srb.Target  = dh->target;
   srb.Lun     = dh->lun;
   switch(data_mode)
   {  case DATA_WRITE:
	srb.Flags = SRB_DIR_OUT | SRB_EVENT_NOTIFY;
	break;
      case DATA_READ:
	srb.Flags = SRB_DIR_IN | SRB_EVENT_NOTIFY;
	break;
      default:
	Stop(_("illegal data_mode for ASPI: %d"),data_mode);
   }
   srb.BufPtr  = buf;
   srb.BufLen  = size;
   srb.SenseLen = 16;
   srb.PostProc = srb_event;

   srb.CDBLen  = cdb_size;
   memcpy(&srb.CDBByte, cmd, cdb_size);

   /* Send the SCSI command */

   ResetEvent(srb_event);
   status = dh->SendASPI32Command(&srb);

   if(status == SS_PENDING)
     WaitForSingleObject(srb_event, INFINITE);

   CloseHandle(srb_event);
   memcpy(sense, &srb.SenseArea, 16);

   /* SS_COMP implies no SCSI error */

   if(srb.Status == SS_COMP) return 0;

   /* Now see what went wrong */

   switch(srb.TargStat)
   {  case 0x00:  /* STATUS_GOOD */
	return -1;

      case 0x02: /* CHECK CONDITION */
	return -1;

      case 0x08: /* BUSY */
	PrintMsg(_("ASPI layer: Target busy.\n"));
	return -1;

      case 0x18: /* Reservation conflict */
	PrintMsg(_("ASPI layer: Reservation conflict.\n"));
	return -1;
   }

   return -1;
}

/**
 ** The SPTI wrapper.
 **/

/* 
 * The Mingw includes are missing the following structs 
 * for sending SCSI packets via SPTI.
 */

typedef struct 
{ USHORT Length;
  UCHAR  ScsiStatus;
  UCHAR  PathId;
  UCHAR  TargetId;
  UCHAR  Lun;
  UCHAR  CdbLength;
  UCHAR  SenseInfoLength;
  UCHAR  DataIn;
  ULONG  DataTransferLength;
  ULONG  TimeOutValue;
  PVOID  DataBuffer;
  ULONG  SenseInfoOffset;
  UCHAR  Cdb[16];
} SCSI_PASS_THROUGH_DIRECT;

typedef struct 
{ SCSI_PASS_THROUGH_DIRECT spt;
  ULONG  Filler;
  UCHAR  SenseBuf[32];
} SCSI_PASS_THROUGH_DIRECT_WITH_BUFFER;

#define IOCTL_SCSI_BASE 0x00000004
#define IOCTL_SCSI_PASS_THROUGH_DIRECT  CTL_CODE( IOCTL_SCSI_BASE, 0x0405, METHOD_BUFFERED, FILE_READ_ACCESS | FILE_WRITE_ACCESS )

/*
 * Open the device using SPTI.
 */

static device_handle* open_spti_device(char *device)
{  device_handle *dh; 
   char device_path[16];

   if(   strlen(device) != 2
      || toupper(*device) < 'C'
      || toupper(*device) > 'Z'
      || device[1] != ':')
	Stop(_("\nIllegal device name \"%s\" (use devices \"C:\" ... \"Z:\")"),device);

   dh = g_malloc0(sizeof(device_handle));
   dh->device = g_strdup(device);

   g_sprintf(device_path, "\\\\.\\%c:", toupper(*device));

   dh->fd = CreateFile(device_path,                  
	               GENERIC_READ | GENERIC_WRITE,
	               FILE_SHARE_READ | FILE_SHARE_WRITE,
                       NULL, OPEN_EXISTING, 0, NULL );

   if(dh->fd == INVALID_HANDLE_VALUE)  /* Might be Win9x or missing priviledges */
   {  g_free(dh);
      return NULL;
   }

   return dh;
}

/*
 * Send the SCSI command through SPTI.
 * At least W2K seems to have problems with SCSI commands timing out;
 * it even manages to do a spontaneous reboot because of that occasionally. 
 * So we set the time out value to a conservative 120 secs.
 */ 

static int send_spti_packet(HANDLE fd, char *cmd, int cdb_size, char *buf, int size, Sense *sense, int data_mode)
{  SCSI_PASS_THROUGH_DIRECT_WITH_BUFFER ss;
   BOOL success;
   ULONG returned;

   memset(&ss, 0, sizeof(ss));

   ss.spt.Length          	= sizeof(SCSI_PASS_THROUGH_DIRECT);
   ss.spt.CdbLength       	= cdb_size;
   ss.spt.SenseInfoLength 	= 24;
   ss.spt.TargetId              = 1;
   switch(data_mode)
   {  case DATA_WRITE:
	ss.spt.DataIn          	= 0;  /* SCSI_IOCTL_DATA_OUT */
	break;
      case DATA_READ:
	ss.spt.DataIn          	= 1;  /* SCSI_IOCTL_DATA_IN */
	break;
      case DATA_NONE:
	ss.spt.DataIn          	= 2;  /* SCSI_IOCTL_DATA_UNSPECIFIED */
	break;
      default:
	Stop(_("illegal data_mode: %d"),data_mode);
   }
   ss.spt.DataTransferLength 	= size;
   ss.spt.TimeOutValue    	= 120;  /* see comment above */
   ss.spt.DataBuffer      	= buf;
   ss.spt.SenseInfoOffset 	= offsetof(SCSI_PASS_THROUGH_DIRECT_WITH_BUFFER, SenseBuf);
   memcpy(ss.spt.Cdb, cmd, cdb_size);

   success = DeviceIoControl(fd, IOCTL_SCSI_PASS_THROUGH_DIRECT,
			    &ss, sizeof(SCSI_PASS_THROUGH_DIRECT_WITH_BUFFER),
			    &ss, sizeof(SCSI_PASS_THROUGH_DIRECT_WITH_BUFFER),
			    &returned, FALSE);

   memcpy(sense, ss.SenseBuf, 24);  /* Hmm. What a kludge. */

   if(!success || ss.spt.ScsiStatus) return -1;
   return 0;
}

/*
 * Dispatch between SPTI and ASPI for packet sending.
 */

static int send_packet(device_handle *dh, char *cmd, int cdb_size, char *buf, int size, Sense *sense, int data_mode)
{
  if(dh->aspiLib) 
       return send_aspi_packet(dh, cmd, cdb_size, buf, size, sense, data_mode);
  else return send_spti_packet(dh->fd, cmd, cdb_size, buf, size, sense, data_mode);
}

#endif /* SYS_MINGW */

/***
 *** CD and DVD query routines.
 *** Everything below should be system independent.
 ***/

/*
 * Send INQUIRY to the device.
 */

static void inquire_device(device_handle *dh)
{  Sense sense;
   char *ibuf;
   char cmd[MAX_CDB_SIZE];
   unsigned char buf[36]; 

   /*** Try to learn something about the device vendor */

   memset(cmd, 0, MAX_CDB_SIZE);
   cmd[0] = 0x12;   /* INQUIRY */
   cmd[4] = 36;     /* allocation length */

   if(send_packet(dh, cmd, 6, buf, 36, &sense, DATA_READ)<0)
   {  strcpy(dh->devinfo, _("unknown"));
#ifdef SYS_MINGW
      Stop(_("\n* INQUIRY failed. Something is wrong with drive %s.\n"),dh->device);
#endif
#ifdef SYS_LINUX
      Stop(_("\n* Can open %s, but INQUIRY fails.\n"
	     "* Chances are that you're using ide-scsi emulation for an ATAPI drive,\n"
	     "* and try to access it via /dev/cdrom or /dev/hd?.\n"
	     "* Either use /dev/scd? or /dev/sr? instead, or disable ide-scsi emulation.\n"),
	   dh->device);
#endif
   }
   else
   {  int i,j; 

      ibuf = dh->devinfo;
      for(i=0,j=8; j<36; j++)   
      {  if(j==32 && i>0 && !isspace(ibuf[i-1])) /* separate the version string */
	   ibuf[i++] = ' ';
         if(   isprint(buf[j])                  /* eliminate multiple spaces and unprintables */
            && (!isspace(buf[j]) || (i>0 && !isspace(ibuf[i-1]))))
	      ibuf[i++] = buf[j];
      }
      ibuf[i] = 0;

      if(buf[0] != 0x05)
      {  if(buf[0]) Stop(_("\nDevice %s (%s) is not a CDROM drive."),dh->device,ibuf);
	 else       Stop(_("\nDevice %s (%s) is a hard disk."),dh->device,ibuf);
      }
   }
}

/*
 * Find out what type of disc has been inserted.
 */

static void query_type(device_handle *dh)
{  static char sbuf[32];
   Sense sense;
   char cmd[MAX_CDB_SIZE];
   unsigned char buf[4096]; 
   unsigned int length;
   int control;
   unsigned int ua_start,ua_end;

   /*** If the medium is a DVD, the following query will succeed. */

   dh->mainType = DVD;  /* default assumption */

   /* Query length of returned data */

   memset(cmd, 0, MAX_CDB_SIZE);
   cmd[0] = 0xad;     /* READ DVD STRUCTURE */
   cmd[6] = 0;        /* First layer */
   cmd[7] = 0;        /* We want PHYSICAL info */
   cmd[8] = 0;        /* Allocation length */
   cmd[9] = 4;

   /* Different CDROMs react with different error codes on this request;
      I have already seen 0x05/0x30/0x02 and 0x05/0x20/0x00 here.
      So we take a conservative route here and test only for the sense
      key 0x05. */

   if(send_packet(dh, cmd, 12, buf, 4, &sense, DATA_READ)<0)
   {  if(sense.sense_key == 0x05)
        goto assume_cd;
    
      PrintSenseNow(sense.sense_key, sense.asc, sense.ascq);
      Stop(_("Could not query dvd structure length: %s\n"),strerror(errno));
   }

   length = buf[0]<<8 | buf[1];
   length += 2;

   if(length>4096) /* don't let the drive hack us using a buffer overflow ;-) */
     Stop(_("Could not query dvd physical structure - implausible packet length %d\n"),length);

   /* Do the real query */

   memset(cmd, 0, MAX_CDB_SIZE);
   cmd[0] = 0xad;     /* READ DVD STRUCTURE */
   cmd[6] = 0;        /* First layer */
   cmd[7] = 0;        /* We want PHYSICAL info */
   cmd[8] = (length>>8) & 0xff;  /* Allocation length */
   cmd[9] = length & 0xff;

   if(send_packet(dh, cmd, 12, buf, length, &sense, DATA_READ)<0)
   {  PrintSenseNow(sense.sense_key, sense.asc, sense.ascq);
      Stop(_("Could not query physical dvd structure: %s\n"),strerror(errno));
   }

   /* See what we've got */

#if 0
   {  int layer_type = buf[6] & 0x0f;
      printf("Layer type(s): ");
      if(layer_type & 0x01) printf("embossed ");
      if(layer_type & 0x02) printf("recordable ");
      if(layer_type & 0x04) printf("rewriteable ");
      printf("\n");
      fflush(stdout);
   }
#endif

   switch((buf[4]>>4) & 0x0f)   /* evaluate the book type */
   {  case  2: dh->typedescr = "DVD-R"; break;
      case  3: dh->typedescr = "DVD-RW"; break;
      case  9: dh->typedescr = "DVD+RW"; break;
      case 10: dh->typedescr = "DVD+R"; break;

      case  0: 
      {  int layer_type = buf[6] & 0x0f;

	 if(layer_type & 0x06) 
	    dh->typedescr = "DVD-ROM (faked book type)";
	 else
	 {  dh->typedescr = "DVD-ROM";
	    dh->subType = UNSUPPORTED;
	    return;
	 }
	 break;
      }

      case  1: 
	dh->typedescr = "DVD-RAM"; 
	dh->subType = UNSUPPORTED; return;
      
      default: 
	g_sprintf(sbuf,"DVD book type 0x%02x",(buf[4]>>4) & 0x0f); 
	dh->typedescr=sbuf; 
	break;
   }

   dh->layers = 1 + ((buf[6] & 0x60) >> 5);
   dh->sessions = 1;
   dh->subType = DVD;

   /* While we're at it, extract the user area size.
      For +RW media, this is better than the value provided by READ CAPACITY. */

   ua_start = buf[ 8]<<24 | buf[ 9]<<16 | buf[10]<<8 | buf[11]; 
   ua_end   = buf[12]<<24 | buf[13]<<16 | buf[14]<<8 | buf[15];
   dh->userAreaSize = (gint64)(ua_end-ua_start);
   return;

   /*** If we reach this, the medium is assumed to be one of the CD varities.*/

assume_cd:
   dh->mainType = CD;

   /*** First, do a READ TOC with format 0 to fetch the CONTROL field. */
   
   memset(cmd, 0, MAX_CDB_SIZE);
   cmd[0] = 0x43;  /* READ TOC/PMA/ATIP */
   cmd[2] = 0;     /* format; we want the TOC */
   cmd[6] = 1;     /* track/session number */
   cmd[7] = 0;     /* allocation length */
   cmd[8] = 2;

   if(send_packet(dh, cmd, 10, buf, 2, &sense, DATA_READ)<0)
   {  PrintSenseNow(sense.sense_key, sense.asc, sense.ascq);
      Stop(_("Could not query TOC length: %s\n"),strerror(errno));
   }

   length = buf[0]<<8 | buf[1];
   length -= 2;  /* MMC3: "Disc information length excludes itself" */
   if(length>1024) /* don't let the drive hack us using a buffer overflow ;-) */
     Stop(_("TOC info too long (%d), probably multisession.\n"),length);

   memset(cmd, 0, MAX_CDB_SIZE);
   cmd[0] = 0x43;  /* READ TOC/PMA/ATIP */
   cmd[2] = 0;     /* format; we want the TOC */
   cmd[6] = 1;     /* track/session number */
   cmd[7] = 0;     /* allocation length */
   cmd[8] = length;

   if(send_packet(dh, cmd, 10, buf, length, &sense, DATA_READ)<0)
   {  PrintSenseNow(sense.sense_key, sense.asc, sense.ascq);
      Stop(_("Could not read TOC: %s\n"),strerror(errno));
   }

   control = buf[5];

   /*** Do the READ TOC again with format 2 to fetch the full toc
        as we want the disc type info also. 
	We do not use the CONTROL data included here as it turned
	out to be invalid on certain CD-ROMs. Bleah. */

   memset(cmd, 0, MAX_CDB_SIZE);
   cmd[0] = 0x43;  /* READ TOC/PMA/ATIP */
   cmd[2] = 2;     /* format; we want the full TOC */
   cmd[6] = 1;     /* track/session number */
   cmd[7] = 0;     /* allocation length */
   cmd[8] = 2;

   if(send_packet(dh, cmd, 10, buf, 2, &sense, DATA_READ)<0)
   {  PrintSenseNow(sense.sense_key, sense.asc, sense.ascq);
      Stop(_("Could not query full TOC length: %s\n"),strerror(errno));
   }

   length = buf[0]<<8 | buf[1];
   length -= 2;  /* MMC3: "Disc information length excludes itself" */
   if(length < 15)
     Stop(_("TOC info too short, length %d.\n"),length);
   if(length>1024) /* don't let the drive hack us using a buffer overflow ;-) */
     Stop(_("TOC info too long (%d), probably multisession.\n"),length);

   memset(cmd, 0, MAX_CDB_SIZE);
   cmd[0] = 0x43;  /* READ TOC/PMA/ATIP */
   cmd[2] = 2;     /* format; we want the full TOC */
   cmd[6] = 1;     /* track/session number */
   cmd[7] = 0;     /* allocation length */
   cmd[8] = length;

   if(send_packet(dh, cmd, 10, buf, length, &sense, DATA_READ)<0)
   {  PrintSenseNow(sense.sense_key, sense.asc, sense.ascq);
      Stop(_("Could not read full TOC: %s\n"),strerror(errno));
   }

   if(buf[7] != 0xa0)
   {  int i;
      PrintMsg(_("\nUnexpected TOC format (length %d):\n"),length);

      for(i=0; i<(int)length; i++) 
      {  PrintMsg("%02x ",buf[i]);
	 if(i==3 || (i-3)%11 == 0) PrintMsg("\n");
      }
      Stop(_("Consider sending a bug report.\n"));
   }

   dh->sessions = buf[3];

   if(control & 4)
     switch(buf[13])
     {  case 0x00: dh->typedescr = _("Data CD mode 1"); dh->subType = DATA1; break;
        case 0x10: dh->typedescr = _("CD-I"); dh->subType = UNSUPPORTED; break;
        case 0x20: dh->typedescr = _("CD XA form 1"); dh->subType = XA21; break;
        default:   dh->typedescr = _("Unknown"); dh->subType = UNSUPPORTED; break;
     }
   else 
   {  dh->typedescr = _("CD Audio"); 
      dh->subType = UNSUPPORTED;
   }
}

/*
 * Find out whether we are allowed to create an image from the DVD.
 */

static int query_copyright(device_handle *dh)
{  Sense sense;
   char cmd[MAX_CDB_SIZE];
   unsigned char buf[4096]; 
   unsigned int length;

   /* Query length of returned data */

   memset(cmd, 0, MAX_CDB_SIZE);
   cmd[0] = 0xad;     /* READ DVD STRUCTURE */
   cmd[6] = 0;        /* First layer */
   cmd[7] = 1;        /* We want copyright info */
   cmd[8] = 0;        /* Allocation length */
   cmd[9] = 4;

   if(send_packet(dh, cmd, 12, buf, 4, &sense, DATA_READ)<0)
   {  PrintSenseNow(sense.sense_key, sense.asc, sense.ascq);
      Stop(_("Could not query dvd structure length for format code 1: %s\n"),strerror(errno));
   }

   length = buf[0]<<8 | buf[1];
   length += 2;

   if(length>4096) /* don't let the drive hack us using a buffer overflow ;-) */
     Stop(_("Could not query dvd copyright info - implausible packet length %d\n"),length);

   /* Do the real query */

   memset(cmd, 0, MAX_CDB_SIZE);
   cmd[0] = 0xad;     /* READ DVD STRUCTURE */
   cmd[6] = 0;        /* First layer */
   cmd[7] = 1;        /* We want copyright info */
   cmd[8] = (length>>8) & 0xff;  /* Allocation length */
   cmd[9] = length & 0xff;

   if(send_packet(dh, cmd, 12, buf, length, &sense, DATA_READ)<0)
   {  PrintSenseNow(sense.sense_key, sense.asc, sense.ascq);
      Stop(_("Could not query copyright info: %s\n"),strerror(errno));
   }

   return buf[4];
}

/*
 * Query the medium size.
 */

static unsigned int query_size(device_handle *dh)
{  Sense sense;
   char cmd[MAX_CDB_SIZE];
   unsigned char buf[8]; 
   unsigned int size;

   memset(cmd, 0, MAX_CDB_SIZE);
   cmd[0] = 0x25;  /* READ CAPACITY */

   if(send_packet(dh, cmd, 10, buf, 8, &sense, DATA_READ)<0)
   {  PrintSenseNow(sense.sense_key, sense.asc, sense.ascq);
      Stop(_("Could not query medium size: %s\n"), strerror(errno));
   }

   size = buf[0]<<24 | buf[1]<<16 | buf[2]<<8 | buf[3];

   /* Rewriteable media return the overall number of available sectors
      which might be larger than the currently formatted area.
      For these types, we use the value provided by READ DVD STRUCTURE
      in the query_type() call. */

   if(dh->subType == DVD && dh->layers == 1)
   {  if(size != dh->userAreaSize 
	 && strcmp(dh->typedescr,"DVD+RW")
	 && strcmp(dh->typedescr,"DVD-RW"))
       PrintMsg(_("* Warning: Different media sizes depending on query method:\n"
	          "*          READ CAPACITY:      %d sectors\n"
	          "*          READ DVD STRUCTURE: %lld sectors\n"
	          "*          This is expected and okay for +RW/-RW media;\n"
		  "*          but your media book type is different.\n"),
	       size+1, dh->userAreaSize+1); 
      return dh->userAreaSize+1;
   }

   /* Use the result of READ CAPACITY for CD and dual layer DVD */

   return size+1;  /* size is the number of the last sector, starting with 0 */
}

/*
 * Set the reading speed. Does not work reliably yet. 
 */

#if 0
static void set_speed(device_handle *dh)
{  Sense sense;
   char cmd[MAX_CDB_SIZE];
   char buf[4];
   int speed_kb;
   int rotcntl;

   if(driveSpeed < 0) 
   {  rotcntl = 1; 
      driveSpeed = -driveSpeed;
   }
   else rotcntl = 0;

   if(!driveSpeed)		/* use maximum performance */
        speed_kb = 0xffff;
   else speed_kb = dh->mainType == DVD ? 1353*driveSpeed : 177*driveSpeed;

   if(speed_kb > 0xffff) speed_kb = 0xffff;

   memset(cmd, 0, MAX_CDB_SIZE);
   cmd[0] = 0xbb;            /* SET CD SPEED */
   cmd[1] = rotcntl & 0x01;  /* CLV or CAV speed */
   cmd[2] = (speed_kb >> 8) & 0xff;
   cmd[3] = speed_kb & 0xff;
   cmd[4] = 0xff;            /* Set write speed always to max */
   cmd[5] = 0xff;

   if(send_packet(dh, cmd, 12, buf, 4, &sense, DATA_READ)<0)
   {  PrintMsg(_("* Warning: Setting the drive speed failed:\n"));
      PrintSenseNow(sense.sense_key, sense.asc, sense.ascq);
   }
}
#endif

/*
 * Sector reading using the packet interface.
 * Using read() hangs some Linux systems on read errors,
 * so we can't go that route.
 */

static int read_dvd_sector(device_handle *dh, char *buf, int lba, int nsectors)
{  Sense *sense = &dh->sense;
   char cmd[MAX_CDB_SIZE];
   int ret;

   memset(cmd, 0, MAX_CDB_SIZE);
   cmd[0] = 0x28;  /* READ(10) */
   cmd[1] = 0;  /* no special flags */
   cmd[2] = (lba >> 24) & 0xff;
   cmd[3] = (lba >> 16) & 0xff;
   cmd[4] = (lba >>  8) & 0xff;
   cmd[5] = lba & 0xff;
   cmd[6] = 0;         /* reserved */
   cmd[7] = 0;         /* number of sectors */
   cmd[8] = nsectors;  /* read 1 sector */

   ret = send_packet(dh, cmd, 10, buf, 2048*nsectors, sense, DATA_READ);

   if(ret<0) RememberSense(sense->sense_key, sense->asc, sense->ascq);

   return ret;
}

static int read_cd_sector(device_handle *dh, char *buf, int lba, int nsectors)
{  Sense *sense = &dh->sense;
   char cmd[MAX_CDB_SIZE];
   int ret;

   memset(cmd, 0, MAX_CDB_SIZE);
   cmd[0]  = 0xbe;         /* READ CD */
   switch(dh->subType)
   {  case DATA1: cmd[1] = 2<<2; break;  /* data mode 1 */
      case XA21:  cmd[1] = 4<<2; break;  /* xa mode 2 form 1 */
   }
   cmd[2]  = (lba >> 24) & 0xff;
   cmd[3]  = (lba >> 16) & 0xff;
   cmd[4]  = (lba >>  8) & 0xff;
   cmd[5]  = lba & 0xff;
   cmd[6]  = 0;        /* number of sectors to read (3 bytes) */
   cmd[7]  = 0;  
   cmd[8]  = nsectors; /* read nsectors */

   cmd[9]  = 0x10;  /* we want the user data only */
   cmd[10] = 0;    /* reserved stuff */
   cmd[11] = 0;    /* no special wishes for the control byte */

   ret = send_packet(dh, cmd, 12, buf, 2048*nsectors, sense, DATA_READ);

   if(ret<0) RememberSense(sense->sense_key, sense->asc, sense->ascq);
	
   return ret;
}

/*
 * Sector reading through the device handle.
 * dh->read dispatches to one the routines above.
 * Certain error conditions have been (empirically) found to be retryable
 * event if the SCSI specs say otherwise,
 * so we check for them and do up to 3 retries for the user's convenience.
 */

int read_sectors(device_handle *dh, char *buf, gint64 s, int nsectors)
{  int retry,status;

   for(retry=3; retry; retry--)
   {  status = dh->read(dh, buf, s, nsectors);
	
      if(status && retry>1 && dh->sense.sense_key == 4 && dh->sense.asc == 8 && dh->sense.ascq == 3)
      {  PrintSense(_("Retrying Sector %lld because of: "), s);
	 PrintLog("\n");
      }
      else break;
   }

   return status;
}

/***
 *** Try reading the medium and create the image and map.
 ***/

ImageInfo* ReadMedium(char *device, char *imagename, char *arg, int scan_mode)
{  struct stat mystat;
   int nsectors; 
   int image = 0;
   gint64 start,end,s,read_marker;
   int last_percent,percent;
   int rereading = 0;
   gint64 read_ok,read_errors;
   ImageInfo *ii;
   char *dashpos = NULL;
   char *bufbase = NULL;
   char *buf = NULL;
   int status,n;
   gint64 last_sector;
   double single_rate = 0.0, speed = 0.0;
   int tao_tail = 0;
   device_handle *dh;
   GTimer *timer = g_timer_new();
	 
   /*** Align the buffer at a 4096 boundary.
	Might be needed by some SCSI drivers. */

   bufbase = g_malloc(32768 + 4096);
   buf = bufbase + (4096 - ((unsigned int)bufbase & 4095));

   /*** Determine medium accessability and size. */

#ifdef SYS_LINUX
   dh = open_device(device);
   inquire_device(dh);
   PrintMsg(_("\nDevice: %s, %s\n"),device,dh->devinfo);
#endif

#ifdef SYS_MINGW
   dh = NULL;
   if(   (*device >= 'c' && *device <= 'z')
      || (*device >= 'C' && *device <= 'Z'))
     dh = open_spti_device(device);
   if(!dh) dh = open_aspi_device(device, 0);
   if(!dh)
     Stop(_("\nNeither SPTI nor ASPI worked opening %s."),device);
   
   inquire_device(dh);
   PrintMsg(_("\nDevice: %s (%s), %s\n"),
	    device, dh->aspiLib ? "ASPI" : "SPTI", dh->devinfo);
#endif

   query_type(dh);

   ii = g_malloc0(sizeof(ImageInfo));
   ii->name    = imagename;
   ii->sectors = query_size(dh);
   ii->size    = 2048*ii->sectors;
   ii->inLast  = 2048;

   read_errors = read_ok = 0;

   switch(dh->subType)
   {  case DVD:
	single_rate = 1352.54;
	dh->read = read_dvd_sector;
	PrintMsg(_("Medium: %s, %lld sectors, %d layer(s)\n\n"),
	         dh->typedescr,ii->sectors,dh->layers);
	break;

      case DATA1:
      case XA21:
	single_rate = 150.0;
	dh->read = read_cd_sector;
	PrintMsg(_("Medium: %s, %lld sectors\n\n"),dh->typedescr,ii->sectors);
	break;

      default:
	Stop(_("Medium type \"%s\" not supported."),dh->typedescr);
   }

   if(dh->mainType == DVD && query_copyright(dh))
   {  g_fprintf(stderr,(_("\n* Can not continue: Encrypted medium.\n\n")));
      fflush(stderr);
      exit(EXIT_FAILURE);
   }

   if(dh->sessions>1)
     Stop(_("Multisession (%d sessions) not supported."),dh->sessions);

   //   set_speed(dh);

   /*** See if user wants to limit the read range. */

   if(arg)
   {  dashpos = strchr(arg,'-');

      if(dashpos)
      {  *dashpos = 0;
         start = atoi(arg);
	 if(strlen(dashpos+1) == 3 && !strncmp(dashpos+1, "end", 3))
	      end = ii->sectors-1;
         else end = atoi(dashpos+1);
      }
      else start = end = atoi(arg);

      if(start>end || start < 0 || end >= ii->sectors)
         Stop(_("Sectors must be in range [0,%lld].\n"),ii->sectors-1);

      PrintMsg(_("Limiting sector range to [%lld,%lld].\n"),start,end);
   }
   else 
   {  start = 0; end = ii->sectors-1;
   }

   /*** See if we already have an image file. */

   if(scan_mode)  /* don't care if in scan mode */
   { PrintMsg(_("SCAN mode - no image file will be altered / written.\n"));
     read_marker = 0;
   } 
   else          
   { int file_flags;

     if(stat(imagename, &mystat) == -1)
     {  /* Start with a fresh file. */

	file_flags = O_RDWR | O_CREAT;
#ifdef HAVE_O_LARGEFILE
	file_flags |= O_LARGEFILE;
#endif
#ifdef SYS_MINGW
	file_flags |= O_BINARY;
#endif
	if((image = open(imagename, file_flags, S_IRWXU)) == -1)
	  Stop(_("Can't open %s:\n%s"),imagename,strerror(errno));

        PrintMsg(_("Creating new %s image.\n"),imagename);
        rereading = 0;
	read_marker = 0;
     }
     else
     {         
        /* Use the existing file as a starting point.
	   Set the read_marker at the end of the file
	   so that the reader looks for "dead_sector" markers
	   and skips already read blocks. */

	file_flags = O_RDWR;
#ifdef HAVE_O_LARGEFILE
	file_flags |= O_LARGEFILE;
#endif
#ifdef SYS_MINGW
	file_flags |= O_BINARY;
#endif
        if((image = open(imagename, file_flags, S_IRWXU)) == -1)
  	  Stop(_("Can't open %s:\n%s"),imagename,strerror(errno));

        rereading = 1;
	read_marker = mystat.st_size / 2048;

	/* If the image is not complete yet, first aim to read the
	   unvisited sectors before trying to re-read the missing ones. */

	if(!arg && read_marker < ii->sectors-1)
	{  PrintMsg(_("Completing image %s. Continuing with sector %lld.\n"),
	            imagename, read_marker);
	   start = read_marker;
	}
	else PrintMsg(_("Completing image %s. Only missing sectors will be read.\n"), imagename);
     }
   }

   /*** If start > read_marker, fill the gap with dead sector markers. */

   if(!scan_mode && start > read_marker)
   {  s = read_marker;

      if(lseek(image, (gint64)(2048*s), SEEK_SET) == -1)
	Stop(_("lseek/fill failed for %s, sector %lld"),imagename,s);

      while(s++ < start)
      {  n = write(image, Closure->deadSector, 2048);
	 if(n != 2048)
	   Stop(_("Write error filling %s:\n%s"),imagename,strerror(errno));
      }
   }

   /*** Read the medium image. */

   last_percent = -1;
   s = last_sector = start;
   g_timer_start(timer);

   while(s<=end)
   {  /*** Read in sector units until the read marker is reached.
	   Also, only switch to 16 sector reading on 32K boundaries
	   -- this matches the internal DVD structure better. 
           In order to treat the 2 read errors at the end of TAO discs correctly,
           we switch back to per sector reading at the end of the medium. */
#if 0
      if(s >= ((ii->sectors - 16 - 2) & ~15))
         PrintLog("Sector %lld read in 1 mode\n",s);
#endif

      if(s<read_marker || s & 15 || s >= ((ii->sectors - 16 - 2) & ~15) ) 
            nsectors = 1;
      else  nsectors = 16;

      if(s+nsectors > end) nsectors = end-s+1;

      /*** Check if the sector has already been read
           and bypass the rest of the loop respectively. 
           Note that s<read_marker implies nsectors==1. */

      if(!scan_mode && s<read_marker)
      {  if(lseek(image, (gint64)(2048*s), SEEK_SET) == -1)
	   Stop(_("lseek/1 failed for %s, sector %lld"),imagename,s);

         n = read(image, buf, 2048);

	 if(n != 2048)
	   Stop(_("unexpected read error in image for sector %lld"),s);

	 if(memcmp(buf, Closure->deadSector, n))
	   goto step_counter;
      }

      /*** Try reading the next <nsectors> sector(s).
      	   Medium Error (3) and Illegal Request (5) may result from 
	   a medium read problem, but other errors are regarded as fatal. */

      status = read_sectors(dh, buf, s, nsectors);

      if(status && dh->sense.sense_key && dh->sense.sense_key != 3 && dh->sense.sense_key != 5)
      {  PrintSense(_("Sector %lld: "), s);  
	 Stop(_("\nCan not recover from above error."));
	 exit(1);
      }

      /*** Process the read error. */

      if(status)
      {  /* If we are reading this part of the image for the first time (= in fast mode),
	    skip a few sectors across the defective area. */

	 if(nsectors > 1)
	 { int nfill;

	   if(nsectors>=Closure->sectorSkip) nfill = nsectors;
	   else
	   {  if(s+Closure->sectorSkip > end) nfill = end-s+1;
	      else nfill = Closure->sectorSkip;
	   }

	   if(!scan_mode)
	   { int i;

	     /* Write nfill of "dead sector" markers so that
		they are tried again in the following iterations / sessions. */

	     if(lseek(image, (gint64)(2048*s), SEEK_SET) == -1)
	       Stop(_("lseek/3 failed for %s, sector %lld"),imagename,s);

	     for(i=0; i<nfill; i++)
	     {  n = write(image, Closure->deadSector, 2048);
	        if(n != 2048)
		  Stop(_("Write error on %s:\n%s"),imagename,strerror(errno));
	     }
	   }

	   /* If sectorSkip is set, perform the skip. 
	      nsectors remains 16 as sectorSkip is a multiple of 16. */

	   if(Closure->sectorSkip)
	   {  read_marker = s + Closure->sectorSkip;
	      nsectors = 16;

	      PrintSense(_("Sector %lld: "), s); 
	      PrintLog(_(" (skipping %d sectors)\n"),nfill-1);  
	      read_errors+=nfill; 
	      s+=nfill-nsectors;   /* nsectors will be added after the goto */
	      goto step_counter;
	   }

           /* Otherwise, slow down for nsectors reads, 
	      and try to re-read the first sector. */

	   else
	   {  read_marker = s + nsectors; 
	      nsectors = 1;
	      status = read_sectors(dh, buf, s, nsectors);

              if(status && dh->sense.sense_key && dh->sense.sense_key != 3 && dh->sense.sense_key != 5)
	      {  PrintSense(_("Sector %lld: "), s);  
		 Stop(_("\nCan not recover from above error."));
	         exit(1);
              }
	   }
	 }

	 if(status)  /* status might have been reset by the re-read above */
	 {  PrintSense(_("Sector %lld: "), s); 
	    PrintLog("\n");
	    if(s >= ii->sectors - 2) tao_tail++;
	    read_errors++;

	    /* mark this sector as missing */

	    if(!scan_mode)
	    {  if(lseek(image, (gint64)(2048*s), SEEK_SET) == -1)
		 Stop(_("lseek/4 failed for %s, sector %lld"),imagename,s);

	       n = write(image, Closure->deadSector, 2048);
	       if(n != 2048)
	         Stop(_("Write error on %s:\n%s"),imagename,strerror(errno));
	    }
	    goto step_counter;
	 }
      }

      /*** Store sector(s) in the image file */

      read_ok += nsectors;

      if(!scan_mode)
      {  if(lseek(image, (gint64)(2048*s), SEEK_SET) == -1)
	   Stop(_("lseek/2 failed for %s, sector %lld: %s"),imagename,s,strerror(errno));

         n = write(image, buf, 2048*nsectors);
	 if(n != 2048*nsectors)
	   Stop(_("Write error on %s:\n%s"),imagename,strerror(errno));
      }

      /*** Step the progress counter */

step_counter:
      s += nsectors;
      percent = (1000*s)/ii->sectors;
      if(last_percent != percent) 
      {  gulong ignore;
	 double kb_read = (s - last_sector) * 2.0;
         double elapsed = g_timer_elapsed(timer, &ignore);
	 double kb_sec  = kb_read / elapsed;

	 if(elapsed > 1.0)  /* sample speed for at least 1sec */
	 {  speed = (speed + kb_sec / single_rate) / 2.0;
	    if(speed>99.9) speed=99.9;

	    last_sector = s;
         }

	 PrintMsg(_("Read position: %3d.%1d%% (%4.1fx)\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b"),
		 percent/10,percent%10,speed);

	 last_percent = percent;

	 if(elapsed > 1.0)
	    g_timer_start(timer);
      }
   }

   if(rereading)
   {  if(!read_errors) PrintMsg(_("\n%lld sectors read.     \n"),read_ok);
      else             PrintMsg(_("\n%lld sectors read; %lld unreadable sectors.\n"),read_ok,read_errors);
   }
   else
   {  if(!read_errors) PrintMsg(_("\nAll sectors successfully read.\n"));
      else             PrintMsg(_("\n%lld unreadable sectors.\n"),read_errors);
   }

   if(dh->mainType != DVD && tao_tail && tao_tail == read_errors && !Closure->noTruncate)
   {  PrintMsg(_("Warning: %d sectors missing at the end of the disc.\n"
                 "         This is okay if the CD was written in TAO (track at once) mode.\n"
		 "         The Image will be trimmed accordingly. See the manual for details.\n"
	         "         Use the --dao option to disable image trimming.\n"),
		 tao_tail);
     
      ii->sectors -= tao_tail;

      if(!scan_mode)
        if(ftruncate(image, (gint64)(2048*ii->sectors)))
	  Stop(_("Could not truncate %s: %s\n"),ii->name,strerror(errno));
   }

   /*** Close and clean up */

   if(!scan_mode) close(image);
   close_device(dh);
   g_free(bufbase);
   g_timer_destroy(timer);

   return ii;
}

